# Copyright (C) 2006-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os

import fastbencode as bencode

from .. import osutils, tests, trace, transform
from .. import revision as _mod_revision
from ..bzr import generate_ids
from ..controldir import ControlDir
from ..errors import (
    StrictCommitFailed,
)
from ..merge import Merge3Merger
from ..mutabletree import MutableTree
from ..osutils import file_kind, pathjoin
from ..transform import (
    ROOT_PARENT,
    MalformedTransform,
    TransformRenameFailed,
    _FileMover,
    resolve_conflicts,
)
from ..transport import FileExists
from . import TestCaseInTempDir, features
from .features import HardlinkFeature, SymlinkFeature


class TransformGroup:
    def __init__(self, dirname, root_id):
        self.name = dirname
        os.mkdir(dirname)
        self.wt = ControlDir.create_standalone_workingtree(dirname)
        if self.wt.supports_file_ids:
            self.wt.set_root_id(root_id)
        self.b = self.wt.branch
        self.tt = self.wt.transform()
        self.root = self.tt.trans_id_tree_path("")


def conflict_text(tree, merge):
    template = b"%s TREE\n%s%s\n%s%s MERGE-SOURCE\n"
    return template % (b"<" * 7, tree, b"=" * 7, merge, b">" * 7)


class TestTransformMerge(TestCaseInTempDir):
    def test_text_merge(self):
        root_id = generate_ids.gen_root_id()
        base = TransformGroup("base", root_id)
        base.tt.new_file("a", base.root, [b"a\nb\nc\nd\be\n"], b"a")
        base.tt.new_file("b", base.root, [b"b1"], b"b")
        base.tt.new_file("c", base.root, [b"c"], b"c")
        base.tt.new_file("d", base.root, [b"d"], b"d")
        base.tt.new_file("e", base.root, [b"e"], b"e")
        base.tt.new_file("f", base.root, [b"f"], b"f")
        base.tt.new_directory("g", base.root, b"g")
        base.tt.new_directory("h", base.root, b"h")
        base.tt.apply()
        other = TransformGroup("other", root_id)
        other.tt.new_file("a", other.root, [b"y\nb\nc\nd\be\n"], b"a")
        other.tt.new_file("b", other.root, [b"b2"], b"b")
        other.tt.new_file("c", other.root, [b"c2"], b"c")
        other.tt.new_file("d", other.root, [b"d"], b"d")
        other.tt.new_file("e", other.root, [b"e2"], b"e")
        other.tt.new_file("f", other.root, [b"f"], b"f")
        other.tt.new_file("g", other.root, [b"g"], b"g")
        other.tt.new_file("h", other.root, [b"h\ni\nj\nk\n"], b"h")
        other.tt.new_file("i", other.root, [b"h\ni\nj\nk\n"], b"i")
        other.tt.apply()
        this = TransformGroup("this", root_id)
        this.tt.new_file("a", this.root, [b"a\nb\nc\nd\bz\n"], b"a")
        this.tt.new_file("b", this.root, [b"b"], b"b")
        this.tt.new_file("c", this.root, [b"c"], b"c")
        this.tt.new_file("d", this.root, [b"d2"], b"d")
        this.tt.new_file("e", this.root, [b"e2"], b"e")
        this.tt.new_file("f", this.root, [b"f"], b"f")
        this.tt.new_file("g", this.root, [b"g"], b"g")
        this.tt.new_file("h", this.root, [b"1\n2\n3\n4\n"], b"h")
        this.tt.new_file("i", this.root, [b"1\n2\n3\n4\n"], b"i")
        this.tt.apply()
        Merge3Merger(this.wt, this.wt, base.wt, other.wt)

        # textual merge
        with this.wt.get_file("a") as f:
            self.assertEqual(f.read(), b"y\nb\nc\nd\bz\n")
        # three-way text conflict
        with this.wt.get_file("b") as f:
            self.assertEqual(f.read(), conflict_text(b"b", b"b2"))
        # OTHER wins
        self.assertEqual(this.wt.get_file("c").read(), b"c2")
        # THIS wins
        self.assertEqual(this.wt.get_file("d").read(), b"d2")
        # Ambigious clean merge
        self.assertEqual(this.wt.get_file("e").read(), b"e2")
        # No change
        self.assertEqual(this.wt.get_file("f").read(), b"f")
        # Correct correct results when THIS == OTHER
        self.assertEqual(this.wt.get_file("g").read(), b"g")
        # Text conflict when THIS & OTHER are text and BASE is dir
        self.assertEqual(
            this.wt.get_file("h").read(),
            conflict_text(b"1\n2\n3\n4\n", b"h\ni\nj\nk\n"),
        )
        self.assertEqual(this.wt.get_file("h.THIS").read(), b"1\n2\n3\n4\n")
        self.assertEqual(this.wt.get_file("h.OTHER").read(), b"h\ni\nj\nk\n")
        self.assertEqual(file_kind(this.wt.abspath("h.BASE")), "directory")
        self.assertEqual(
            this.wt.get_file("i").read(),
            conflict_text(b"1\n2\n3\n4\n", b"h\ni\nj\nk\n"),
        )
        self.assertEqual(this.wt.get_file("i.THIS").read(), b"1\n2\n3\n4\n")
        self.assertEqual(this.wt.get_file("i.OTHER").read(), b"h\ni\nj\nk\n")
        self.assertEqual(os.path.exists(this.wt.abspath("i.BASE")), False)
        modified = ["a", "b", "c", "h", "i"]
        merge_modified = this.wt.merge_modified()
        self.assertSubset(merge_modified, modified)
        self.assertEqual(len(merge_modified), len(modified))
        with open(this.wt.abspath("a"), "wb") as f:
            f.write(b"booga")
        modified.pop(0)
        merge_modified = this.wt.merge_modified()
        self.assertSubset(merge_modified, modified)
        self.assertEqual(len(merge_modified), len(modified))
        this.wt.remove("b")
        this.wt.revert()

    def test_file_merge(self):
        self.requireFeature(SymlinkFeature(self.test_dir))
        root_id = generate_ids.gen_root_id()
        base = TransformGroup("BASE", root_id)
        this = TransformGroup("THIS", root_id)
        other = TransformGroup("OTHER", root_id)
        for tg in this, base, other:
            tg.tt.new_directory("a", tg.root, b"a")
            tg.tt.new_symlink("b", tg.root, "b", b"b")
            tg.tt.new_file("c", tg.root, [b"c"], b"c")
            tg.tt.new_symlink("d", tg.root, tg.name, b"d")
        targets = (
            (base, "base-e", "base-f", None, None),
            (this, "other-e", "this-f", "other-g", "this-h"),
            (other, "other-e", None, "other-g", "other-h"),
        )
        for tg, e_target, f_target, g_target, h_target in targets:
            for link, target in (
                ("e", e_target),
                ("f", f_target),
                ("g", g_target),
                ("h", h_target),
            ):
                if target is not None:
                    tg.tt.new_symlink(link, tg.root, target, link.encode("ascii"))

        for tg in this, base, other:
            tg.tt.apply()
        Merge3Merger(this.wt, this.wt, base.wt, other.wt)
        self.assertIs(os.path.isdir(this.wt.abspath("a")), True)
        self.assertIs(os.path.islink(this.wt.abspath("b")), True)
        self.assertIs(os.path.isfile(this.wt.abspath("c")), True)
        for suffix in ("THIS", "BASE", "OTHER"):
            self.assertEqual(os.readlink(this.wt.abspath("d." + suffix)), suffix)
        self.assertIs(os.path.lexists(this.wt.abspath("d")), False)
        self.assertEqual(this.wt.id2path(b"d"), "d.OTHER")
        self.assertEqual(this.wt.id2path(b"f"), "f.THIS")
        self.assertEqual(os.readlink(this.wt.abspath("e")), "other-e")
        self.assertIs(os.path.lexists(this.wt.abspath("e.THIS")), False)
        self.assertIs(os.path.lexists(this.wt.abspath("e.OTHER")), False)
        self.assertIs(os.path.lexists(this.wt.abspath("e.BASE")), False)
        self.assertIs(os.path.lexists(this.wt.abspath("g")), True)
        self.assertIs(os.path.lexists(this.wt.abspath("g.BASE")), False)
        self.assertIs(os.path.lexists(this.wt.abspath("h")), False)
        self.assertIs(os.path.lexists(this.wt.abspath("h.BASE")), False)
        self.assertIs(os.path.lexists(this.wt.abspath("h.THIS")), True)
        self.assertIs(os.path.lexists(this.wt.abspath("h.OTHER")), True)

    def test_filename_merge(self):
        root_id = generate_ids.gen_root_id()
        base = TransformGroup("BASE", root_id)
        this = TransformGroup("THIS", root_id)
        other = TransformGroup("OTHER", root_id)
        base_a, this_a, other_a = (
            t.tt.new_directory("a", t.root, b"a") for t in [base, this, other]
        )
        base_b, this_b, other_b = (
            t.tt.new_directory("b", t.root, b"b") for t in [base, this, other]
        )
        base.tt.new_directory("c", base_a, b"c")
        this.tt.new_directory("c1", this_a, b"c")
        other.tt.new_directory("c", other_b, b"c")

        base.tt.new_directory("d", base_a, b"d")
        this.tt.new_directory("d1", this_b, b"d")
        other.tt.new_directory("d", other_a, b"d")

        base.tt.new_directory("e", base_a, b"e")
        this.tt.new_directory("e", this_a, b"e")
        other.tt.new_directory("e1", other_b, b"e")

        base.tt.new_directory("f", base_a, b"f")
        this.tt.new_directory("f1", this_b, b"f")
        other.tt.new_directory("f1", other_b, b"f")

        for tg in [this, base, other]:
            tg.tt.apply()
        Merge3Merger(this.wt, this.wt, base.wt, other.wt)
        self.assertEqual(this.wt.id2path(b"c"), pathjoin("b/c1"))
        self.assertEqual(this.wt.id2path(b"d"), pathjoin("b/d1"))
        self.assertEqual(this.wt.id2path(b"e"), pathjoin("b/e1"))
        self.assertEqual(this.wt.id2path(b"f"), pathjoin("b/f1"))

    def test_filename_merge_conflicts(self):
        root_id = generate_ids.gen_root_id()
        base = TransformGroup("BASE", root_id)
        this = TransformGroup("THIS", root_id)
        other = TransformGroup("OTHER", root_id)
        base_a, this_a, other_a = (
            t.tt.new_directory("a", t.root, b"a") for t in [base, this, other]
        )
        base_b, this_b, other_b = (
            t.tt.new_directory("b", t.root, b"b") for t in [base, this, other]
        )

        base.tt.new_file("g", base_a, [b"g"], b"g")
        other.tt.new_file("g1", other_b, [b"g1"], b"g")

        base.tt.new_file("h", base_a, [b"h"], b"h")
        this.tt.new_file("h1", this_b, [b"h1"], b"h")

        base.tt.new_file("i", base.root, [b"i"], b"i")
        other.tt.new_directory("i1", this_b, b"i")

        for tg in [this, base, other]:
            tg.tt.apply()
        Merge3Merger(this.wt, this.wt, base.wt, other.wt)

        self.assertEqual(this.wt.id2path(b"g"), pathjoin("b/g1.OTHER"))
        self.assertIs(os.path.lexists(this.wt.abspath("b/g1.BASE")), True)
        self.assertIs(os.path.lexists(this.wt.abspath("b/g1.THIS")), False)
        self.assertEqual(this.wt.id2path(b"h"), pathjoin("b/h1.THIS"))
        self.assertIs(os.path.lexists(this.wt.abspath("b/h1.BASE")), True)
        self.assertIs(os.path.lexists(this.wt.abspath("b/h1.OTHER")), False)
        self.assertEqual(this.wt.id2path(b"i"), pathjoin("b/i1.OTHER"))


class TestCommitTransform(tests.TestCaseWithTransport):
    def get_branch(self):
        tree = self.make_branch_and_tree("tree")
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.commit("empty commit")
        return tree.branch

    def get_branch_and_transform(self):
        branch = self.get_branch()
        tt = branch.basis_tree().preview_transform()
        self.addCleanup(tt.finalize)
        return branch, tt

    def test_commit_wrong_basis(self):
        branch = self.get_branch()
        basis = branch.repository.revision_tree(_mod_revision.NULL_REVISION)
        tt = basis.preview_transform()
        self.addCleanup(tt.finalize)
        e = self.assertRaises(ValueError, tt.commit, branch, "")
        self.assertEqual("TreeTransform not based on branch basis: null:", str(e))

    def test_empy_commit(self):
        branch, tt = self.get_branch_and_transform()
        rev = tt.commit(branch, "my message")
        self.assertEqual(2, branch.revno())
        repo = branch.repository
        self.assertEqual("my message", repo.get_revision(rev).message)

    def test_merge_parents(self):
        branch, tt = self.get_branch_and_transform()
        tt.commit(branch, "my message", [b"rev1b", b"rev1c"])
        self.assertEqual([b"rev1b", b"rev1c"], branch.basis_tree().get_parent_ids()[1:])

    def test_first_commit(self):
        branch = self.make_branch("branch")
        branch.lock_write()
        self.addCleanup(branch.unlock)
        tt = branch.basis_tree().preview_transform()
        self.addCleanup(tt.finalize)
        tt.new_directory("", ROOT_PARENT, b"TREE_ROOT")
        tt.commit(branch, "my message")
        self.assertEqual([], branch.basis_tree().get_parent_ids())
        self.assertNotEqual(_mod_revision.NULL_REVISION, branch.last_revision())

    def test_first_commit_with_merge_parents(self):
        branch = self.make_branch("branch")
        branch.lock_write()
        self.addCleanup(branch.unlock)
        tt = branch.basis_tree().preview_transform()
        self.addCleanup(tt.finalize)
        e = self.assertRaises(
            ValueError, tt.commit, branch, "my message", [b"rev1b-id"]
        )
        self.assertEqual("Cannot supply merge parents for first commit.", str(e))
        self.assertEqual(_mod_revision.NULL_REVISION, branch.last_revision())

    def test_add_files(self):
        branch, tt = self.get_branch_and_transform()
        tt.new_file("file", tt.root, [b"contents"], b"file-id")
        trans_id = tt.new_directory("dir", tt.root, b"dir-id")
        if SymlinkFeature(self.test_dir).available():
            tt.new_symlink("symlink", trans_id, "target", b"symlink-id")
        tt.commit(branch, "message")
        tree = branch.basis_tree()
        self.assertEqual("file", tree.id2path(b"file-id"))
        self.assertEqual(b"contents", tree.get_file_text("file"))
        self.assertEqual("dir", tree.id2path(b"dir-id"))
        if SymlinkFeature(self.test_dir).available():
            self.assertEqual("dir/symlink", tree.id2path(b"symlink-id"))
            self.assertEqual("target", tree.get_symlink_target("dir/symlink"))

    def test_add_unversioned(self):
        branch, tt = self.get_branch_and_transform()
        tt.new_file("file", tt.root, [b"contents"])
        self.assertRaises(StrictCommitFailed, tt.commit, branch, "message", strict=True)

    def test_modify_strict(self):
        branch, tt = self.get_branch_and_transform()
        tt.new_file("file", tt.root, [b"contents"], b"file-id")
        tt.commit(branch, "message", strict=True)
        tt = branch.basis_tree().preview_transform()
        self.addCleanup(tt.finalize)
        trans_id = tt.trans_id_file_id(b"file-id")
        tt.delete_contents(trans_id)
        tt.create_file([b"contents"], trans_id)
        tt.commit(branch, "message", strict=True)

    def test_commit_malformed(self):
        """Committing a malformed transform should raise an exception.

        In this case, we are adding a file without adding its parent.
        """
        branch, tt = self.get_branch_and_transform()
        parent_id = tt.trans_id_file_id(b"parent-id")
        tt.new_file("file", parent_id, [b"contents"], b"file-id")
        self.assertRaises(MalformedTransform, tt.commit, branch, "message")

    def test_commit_rich_revision_data(self):
        branch, tt = self.get_branch_and_transform()
        rev_id = tt.commit(
            branch,
            "message",
            timestamp=1,
            timezone=43201,
            committer="me <me@example.com>",
            revprops={"foo": "bar"},
            revision_id=b"revid-1",
            authors=[
                "Author1 <author1@example.com>",
                "Author2 <author2@example.com>",
            ],
        )
        self.assertEqual(b"revid-1", rev_id)
        revision = branch.repository.get_revision(rev_id)
        self.assertEqual(1, revision.timestamp)
        self.assertEqual(43201, revision.timezone)
        self.assertEqual("me <me@example.com>", revision.committer)
        self.assertEqual(
            ["Author1 <author1@example.com>", "Author2 <author2@example.com>"],
            revision.get_apparent_authors(),
        )
        del revision.properties["authors"]
        self.assertEqual({"foo": "bar", "branch-nick": "tree"}, revision.properties)

    def test_no_explicit_revprops(self):
        branch, tt = self.get_branch_and_transform()
        rev_id = tt.commit(
            branch,
            "message",
            authors=[
                "Author1 <author1@example.com>",
                "Author2 <author2@example.com>",
            ],
        )
        revision = branch.repository.get_revision(rev_id)
        self.assertEqual(
            ["Author1 <author1@example.com>", "Author2 <author2@example.com>"],
            revision.get_apparent_authors(),
        )
        self.assertEqual("tree", revision.properties["branch-nick"])


class TestFileMover(tests.TestCaseWithTransport):
    def test_file_mover(self):
        self.build_tree(["a/", "a/b", "c/", "c/d"])
        mover = _FileMover()
        mover.rename("a", "q")
        self.assertPathExists("q")
        self.assertPathDoesNotExist("a")
        self.assertPathExists("q/b")
        self.assertPathExists("c")
        self.assertPathExists("c/d")

    def test_pre_delete_rollback(self):
        self.build_tree(["a/"])
        mover = _FileMover()
        mover.pre_delete("a", "q")
        self.assertPathExists("q")
        self.assertPathDoesNotExist("a")
        mover.rollback()
        self.assertPathDoesNotExist("q")
        self.assertPathExists("a")

    def test_apply_deletions(self):
        self.build_tree(["a/", "b/"])
        mover = _FileMover()
        mover.pre_delete("a", "q")
        mover.pre_delete("b", "r")
        self.assertPathExists("q")
        self.assertPathExists("r")
        self.assertPathDoesNotExist("a")
        self.assertPathDoesNotExist("b")
        mover.apply_deletions()
        self.assertPathDoesNotExist("q")
        self.assertPathDoesNotExist("r")
        self.assertPathDoesNotExist("a")
        self.assertPathDoesNotExist("b")

    def test_file_mover_rollback(self):
        self.build_tree(["a/", "a/b", "c/", "c/d/", "c/e/"])
        mover = _FileMover()
        mover.rename("c/d", "c/f")
        mover.rename("c/e", "c/d")
        try:
            mover.rename("a", "c")
        except FileExists:
            mover.rollback()
        self.assertPathExists("a")
        self.assertPathExists("c/d")


class Bogus(Exception):
    pass


class TestTransformRollback(tests.TestCaseWithTransport):
    class ExceptionFileMover(_FileMover):
        def __init__(self, bad_source=None, bad_target=None):
            _FileMover.__init__(self)
            self.bad_source = bad_source
            self.bad_target = bad_target

        def rename(self, source, target):
            if self.bad_source is not None and source.endswith(self.bad_source):
                raise Bogus
            elif self.bad_target is not None and target.endswith(self.bad_target):
                raise Bogus
            else:
                _FileMover.rename(self, source, target)

    def test_rollback_rename(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b"])
        tt = tree.transform()
        self.addCleanup(tt.finalize)
        a_id = tt.trans_id_tree_path("a")
        tt.adjust_path("c", tt.root, a_id)
        tt.adjust_path("d", a_id, tt.trans_id_tree_path("a/b"))
        self.assertRaises(
            Bogus, tt.apply, _mover=self.ExceptionFileMover(bad_source="a")
        )
        self.assertPathExists("a")
        self.assertPathExists("a/b")
        tt.apply()
        self.assertPathExists("c")
        self.assertPathExists("c/d")

    def test_rollback_rename_into_place(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b"])
        tt = tree.transform()
        self.addCleanup(tt.finalize)
        a_id = tt.trans_id_tree_path("a")
        tt.adjust_path("c", tt.root, a_id)
        tt.adjust_path("d", a_id, tt.trans_id_tree_path("a/b"))
        self.assertRaises(
            Bogus, tt.apply, _mover=self.ExceptionFileMover(bad_target="c/d")
        )
        self.assertPathExists("a")
        self.assertPathExists("a/b")
        tt.apply()
        self.assertPathExists("c")
        self.assertPathExists("c/d")

    def test_rollback_deletion(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b"])
        tt = tree.transform()
        self.addCleanup(tt.finalize)
        a_id = tt.trans_id_tree_path("a")
        tt.delete_contents(a_id)
        tt.adjust_path("d", tt.root, tt.trans_id_tree_path("a/b"))
        self.assertRaises(
            Bogus, tt.apply, _mover=self.ExceptionFileMover(bad_target="d")
        )
        self.assertPathExists("a")
        self.assertPathExists("a/b")


class TestFinalizeRobustness(tests.TestCaseWithTransport):
    """Ensure treetransform creation errors can be safely cleaned up after."""

    def _override_globals_in_method(self, instance, method_name, globals):
        """Replace method on instance with one with updated globals."""
        import types

        func = getattr(instance, method_name).__func__
        new_globals = dict(func.__globals__)
        new_globals.update(globals)
        new_func = types.FunctionType(
            func.__code__, new_globals, func.__name__, func.__defaults__
        )
        setattr(instance, method_name, types.MethodType(new_func, instance))
        self.addCleanup(delattr, instance, method_name)

    @staticmethod
    def _fake_open_raises_before(name, mode):
        """Like open() but raises before doing anything."""
        raise RuntimeError

    @staticmethod
    def _fake_open_raises_after(name, mode):
        """Like open() but raises after creating file without returning."""
        open(name, mode).close()
        raise RuntimeError

    def create_transform_and_root_trans_id(self):
        """Setup a transform creating a file in limbo."""
        tree = self.make_branch_and_tree(".")
        tt = tree.transform()
        return tt, tt.create_path("a", tt.root)

    def create_transform_and_subdir_trans_id(self):
        """Setup a transform creating a directory containing a file in limbo."""
        tree = self.make_branch_and_tree(".")
        tt = tree.transform()
        d_trans_id = tt.create_path("d", tt.root)
        tt.create_directory(d_trans_id)
        f_trans_id = tt.create_path("a", d_trans_id)
        tt.adjust_path("a", d_trans_id, f_trans_id)
        return tt, f_trans_id

    def test_root_create_file_open_raises_before_creation(self):
        tt, trans_id = self.create_transform_and_root_trans_id()
        self._override_globals_in_method(
            tt, "create_file", {"open": self._fake_open_raises_before}
        )
        self.assertRaises(RuntimeError, tt.create_file, [b"contents"], trans_id)
        path = tt._limbo_name(trans_id)
        self.assertPathDoesNotExist(path)
        tt.finalize()
        self.assertPathDoesNotExist(tt._limbodir)

    def test_root_create_file_open_raises_after_creation(self):
        tt, trans_id = self.create_transform_and_root_trans_id()
        self._override_globals_in_method(
            tt, "create_file", {"open": self._fake_open_raises_after}
        )
        self.assertRaises(RuntimeError, tt.create_file, [b"contents"], trans_id)
        path = tt._limbo_name(trans_id)
        self.assertPathExists(path)
        tt.finalize()
        self.assertPathDoesNotExist(path)
        self.assertPathDoesNotExist(tt._limbodir)

    def test_subdir_create_file_open_raises_before_creation(self):
        tt, trans_id = self.create_transform_and_subdir_trans_id()
        self._override_globals_in_method(
            tt, "create_file", {"open": self._fake_open_raises_before}
        )
        self.assertRaises(RuntimeError, tt.create_file, [b"contents"], trans_id)
        path = tt._limbo_name(trans_id)
        self.assertPathDoesNotExist(path)
        tt.finalize()
        self.assertPathDoesNotExist(tt._limbodir)

    def test_subdir_create_file_open_raises_after_creation(self):
        tt, trans_id = self.create_transform_and_subdir_trans_id()
        self._override_globals_in_method(
            tt, "create_file", {"open": self._fake_open_raises_after}
        )
        self.assertRaises(RuntimeError, tt.create_file, [b"contents"], trans_id)
        path = tt._limbo_name(trans_id)
        self.assertPathExists(path)
        tt.finalize()
        self.assertPathDoesNotExist(path)
        self.assertPathDoesNotExist(tt._limbodir)

    def test_rename_in_limbo_rename_raises_after_rename(self):
        tt, trans_id = self.create_transform_and_root_trans_id()
        parent1 = tt.new_directory("parent1", tt.root)
        child1 = tt.new_file("child1", parent1, [b"contents"])
        parent2 = tt.new_directory("parent2", tt.root)

        class FakeOSModule:
            def rename(self, old, new):
                os.rename(old, new)
                raise RuntimeError

        self._override_globals_in_method(tt, "_rename_in_limbo", {"os": FakeOSModule()})
        self.assertRaises(RuntimeError, tt.adjust_path, "child1", parent2, child1)
        path = osutils.pathjoin(tt._limbo_name(parent2), "child1")
        self.assertPathExists(path)
        tt.finalize()
        self.assertPathDoesNotExist(path)
        self.assertPathDoesNotExist(tt._limbodir)

    def test_rename_in_limbo_rename_raises_before_rename(self):
        tt, trans_id = self.create_transform_and_root_trans_id()
        parent1 = tt.new_directory("parent1", tt.root)
        child1 = tt.new_file("child1", parent1, [b"contents"])
        parent2 = tt.new_directory("parent2", tt.root)

        class FakeOSModule:
            def rename(self, old, new):
                raise RuntimeError

        self._override_globals_in_method(tt, "_rename_in_limbo", {"os": FakeOSModule()})
        self.assertRaises(RuntimeError, tt.adjust_path, "child1", parent2, child1)
        path = osutils.pathjoin(tt._limbo_name(parent1), "child1")
        self.assertPathExists(path)
        tt.finalize()
        self.assertPathDoesNotExist(path)
        self.assertPathDoesNotExist(tt._limbodir)


class TestTransformMissingParent(tests.TestCaseWithTransport):
    def make_tt_with_versioned_dir(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(
            [
                "dir/",
            ]
        )
        wt.add(["dir"], ids=[b"dir-id"])
        wt.commit("Create dir")
        tt = wt.transform()
        self.addCleanup(tt.finalize)
        return wt, tt

    def test_resolve_create_parent_for_versioned_file(self):
        wt, tt = self.make_tt_with_versioned_dir()
        dir_tid = tt.trans_id_tree_path("dir")
        tt.new_file("file", dir_tid, [b"Contents"], file_id=b"file-id")
        tt.delete_contents(dir_tid)
        tt.unversion_file(dir_tid)
        conflicts = resolve_conflicts(tt)
        # one conflict for the missing directory, one for the unversioned
        # parent
        self.assertLength(2, conflicts)

    def test_non_versioned_file_create_conflict(self):
        wt, tt = self.make_tt_with_versioned_dir()
        dir_tid = tt.trans_id_tree_path("dir")
        tt.new_file("file", dir_tid, [b"Contents"])
        tt.delete_contents(dir_tid)
        tt.unversion_file(dir_tid)
        conflicts = resolve_conflicts(tt)
        # no conflicts or rather: orphaning 'file' resolve the 'dir' conflict
        self.assertLength(1, conflicts)
        self.assertEqual(("deleting parent", "Not deleting", "new-1"), conflicts.pop())


class FakeSerializer:
    """Serializer implementation that simply returns the input.

    The input is returned in the order used by pack.ContainerPushParser.
    """

    @staticmethod
    def bytes_record(bytes, names):
        return names, bytes


class TestSerializeTransform(tests.TestCaseWithTransport):
    _test_needs_features = [features.UnicodeFilenameFeature]

    def get_preview(self, tree=None):
        if tree is None:
            tree = self.make_branch_and_tree("tree")
        tt = tree.preview_transform()
        self.addCleanup(tt.finalize)
        return tt

    def assertSerializesTo(self, expected, tt):
        records = list(tt.serialize(FakeSerializer()))
        self.assertEqual(expected, records)

    @staticmethod
    def default_attribs():
        return {
            b"_id_number": 1,
            b"_new_name": {},
            b"_new_parent": {},
            b"_new_executability": {},
            b"_new_id": {},
            b"_tree_path_ids": {b"": b"new-0"},
            b"_removed_id": [],
            b"_removed_contents": [],
            b"_non_present_ids": {},
        }

    def make_records(self, attribs, contents):
        records = [((((b"attribs"),),), bencode.bencode(attribs))]
        records.extend([(((n, k),), c) for n, k, c in contents])
        return records

    def creation_records(self):
        attribs = self.default_attribs()
        attribs[b"_id_number"] = 3
        attribs[b"_new_name"] = {b"new-1": "foo\u1234".encode(), b"new-2": b"qux"}
        attribs[b"_new_id"] = {b"new-1": b"baz", b"new-2": b"quxx"}
        attribs[b"_new_parent"] = {b"new-1": b"new-0", b"new-2": b"new-0"}
        attribs[b"_new_executability"] = {b"new-1": 1}
        contents = [
            (b"new-1", b"file", b"i 1\nbar\n"),
            (b"new-2", b"directory", b""),
        ]
        return self.make_records(attribs, contents)

    def test_serialize_creation(self):
        tt = self.get_preview()
        tt.new_file("foo\u1234", tt.root, [b"bar"], b"baz", True)
        tt.new_directory("qux", tt.root, b"quxx")
        self.assertSerializesTo(self.creation_records(), tt)

    def test_deserialize_creation(self):
        tt = self.get_preview()
        tt.deserialize(iter(self.creation_records()))
        self.assertEqual(3, tt._id_number)
        self.assertEqual({"new-1": "foo\u1234", "new-2": "qux"}, tt._new_name)
        self.assertEqual({"new-1": b"baz", "new-2": b"quxx"}, tt._new_id)
        self.assertEqual({"new-1": tt.root, "new-2": tt.root}, tt._new_parent)
        self.assertEqual({b"baz": "new-1", b"quxx": "new-2"}, tt._r_new_id)
        self.assertEqual({"new-1": True}, tt._new_executability)
        self.assertEqual({"new-1": "file", "new-2": "directory"}, tt._new_contents)
        with open(tt._limbo_name("new-1"), "rb") as foo_limbo:
            foo_content = foo_limbo.read()
        self.assertEqual(b"bar", foo_content)

    def symlink_creation_records(self):
        attribs = self.default_attribs()
        attribs[b"_id_number"] = 2
        attribs[b"_new_name"] = {b"new-1": "foo\u1234".encode()}
        attribs[b"_new_parent"] = {b"new-1": b"new-0"}
        contents = [(b"new-1", b"symlink", "bar\u1234".encode())]
        return self.make_records(attribs, contents)

    def test_serialize_symlink_creation(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tt = self.get_preview()
        tt.new_symlink("foo\u1234", tt.root, "bar\u1234")
        self.assertSerializesTo(self.symlink_creation_records(), tt)

    def test_deserialize_symlink_creation(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tt = self.get_preview()
        tt.deserialize(iter(self.symlink_creation_records()))
        abspath = tt._limbo_name("new-1")
        foo_content = osutils.readlink(abspath)
        self.assertEqual("bar\u1234", foo_content)

    def make_destruction_preview(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["foo\u1234", "bar"])
        tree.add(["foo\u1234", "bar"], ids=[b"foo-id", b"bar-id"])
        return self.get_preview(tree)

    def destruction_records(self):
        attribs = self.default_attribs()
        attribs[b"_id_number"] = 3
        attribs[b"_removed_id"] = [b"new-1"]
        attribs[b"_removed_contents"] = [b"new-2"]
        attribs[b"_tree_path_ids"] = {
            b"": b"new-0",
            "foo\u1234".encode(): b"new-1",
            b"bar": b"new-2",
        }
        return self.make_records(attribs, [])

    def test_serialize_destruction(self):
        tt = self.make_destruction_preview()
        foo_trans_id = tt.trans_id_tree_path("foo\u1234")
        tt.unversion_file(foo_trans_id)
        bar_trans_id = tt.trans_id_tree_path("bar")
        tt.delete_contents(bar_trans_id)
        self.assertSerializesTo(self.destruction_records(), tt)

    def test_deserialize_destruction(self):
        tt = self.make_destruction_preview()
        tt.deserialize(iter(self.destruction_records()))
        self.assertEqual(
            {"foo\u1234": "new-1", "bar": "new-2", "": tt.root}, tt._tree_path_ids
        )
        self.assertEqual(
            {"new-1": "foo\u1234", "new-2": "bar", tt.root: ""}, tt._tree_id_paths
        )
        self.assertEqual({"new-1"}, tt._removed_id)
        self.assertEqual({"new-2"}, tt._removed_contents)

    def missing_records(self):
        attribs = self.default_attribs()
        attribs[b"_id_number"] = 2
        attribs[b"_non_present_ids"] = {
            b"boo": b"new-1",
        }
        return self.make_records(attribs, [])

    def test_serialize_missing(self):
        tt = self.get_preview()
        tt.trans_id_file_id(b"boo")
        self.assertSerializesTo(self.missing_records(), tt)

    def test_deserialize_missing(self):
        tt = self.get_preview()
        tt.deserialize(iter(self.missing_records()))
        self.assertEqual({b"boo": "new-1"}, tt._non_present_ids)

    def make_modification_preview(self):
        LINES_ONE = b"aa\nbb\ncc\ndd\n"
        LINES_TWO = b"z\nbb\nx\ndd\n"
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/file", LINES_ONE)])
        tree.add("file", ids=b"file-id")
        return self.get_preview(tree), [LINES_TWO]

    def modification_records(self):
        attribs = self.default_attribs()
        attribs[b"_id_number"] = 2
        attribs[b"_tree_path_ids"] = {
            b"file": b"new-1",
            b"": b"new-0",
        }
        attribs[b"_removed_contents"] = [b"new-1"]
        contents = [(b"new-1", b"file", b"i 1\nz\n\nc 0 1 1 1\ni 1\nx\n\nc 0 3 3 1\n")]
        return self.make_records(attribs, contents)

    def test_serialize_modification(self):
        tt, LINES = self.make_modification_preview()
        trans_id = tt.trans_id_file_id(b"file-id")
        tt.delete_contents(trans_id)
        tt.create_file(LINES, trans_id)
        self.assertSerializesTo(self.modification_records(), tt)

    def test_deserialize_modification(self):
        tt, LINES = self.make_modification_preview()
        tt.deserialize(iter(self.modification_records()))
        self.assertFileEqual(b"".join(LINES), tt._limbo_name("new-1"))

    def make_kind_change_preview(self):
        LINES = b"a\nb\nc\nd\n"
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/foo/"])
        tree.add("foo", ids=b"foo-id")
        return self.get_preview(tree), [LINES]

    def kind_change_records(self):
        attribs = self.default_attribs()
        attribs[b"_id_number"] = 2
        attribs[b"_tree_path_ids"] = {
            b"foo": b"new-1",
            b"": b"new-0",
        }
        attribs[b"_removed_contents"] = [b"new-1"]
        contents = [(b"new-1", b"file", b"i 4\na\nb\nc\nd\n\n")]
        return self.make_records(attribs, contents)

    def test_serialize_kind_change(self):
        tt, LINES = self.make_kind_change_preview()
        trans_id = tt.trans_id_file_id(b"foo-id")
        tt.delete_contents(trans_id)
        tt.create_file(LINES, trans_id)
        self.assertSerializesTo(self.kind_change_records(), tt)

    def test_deserialize_kind_change(self):
        tt, LINES = self.make_kind_change_preview()
        tt.deserialize(iter(self.kind_change_records()))
        self.assertFileEqual(b"".join(LINES), tt._limbo_name("new-1"))

    def make_add_contents_preview(self):
        LINES = b"a\nb\nc\nd\n"
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/foo"])
        tree.add("foo")
        os.unlink("tree/foo")
        return self.get_preview(tree), LINES

    def add_contents_records(self):
        attribs = self.default_attribs()
        attribs[b"_id_number"] = 2
        attribs[b"_tree_path_ids"] = {
            b"foo": b"new-1",
            b"": b"new-0",
        }
        contents = [(b"new-1", b"file", b"i 4\na\nb\nc\nd\n\n")]
        return self.make_records(attribs, contents)

    def test_serialize_add_contents(self):
        tt, LINES = self.make_add_contents_preview()
        trans_id = tt.trans_id_tree_path("foo")
        tt.create_file([LINES], trans_id)
        self.assertSerializesTo(self.add_contents_records(), tt)

    def test_deserialize_add_contents(self):
        tt, LINES = self.make_add_contents_preview()
        tt.deserialize(iter(self.add_contents_records()))
        self.assertFileEqual(LINES, tt._limbo_name("new-1"))

    def test_get_parents_lines(self):
        LINES_ONE = b"aa\nbb\ncc\ndd\n"
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/file", LINES_ONE)])
        tree.add("file", ids=b"file-id")
        tt = self.get_preview(tree)
        trans_id = tt.trans_id_tree_path("file")
        self.assertEqual(
            ([b"aa\n", b"bb\n", b"cc\n", b"dd\n"],), tt._get_parents_lines(trans_id)
        )

    def test_get_parents_texts(self):
        LINES_ONE = b"aa\nbb\ncc\ndd\n"
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/file", LINES_ONE)])
        tree.add("file", ids=b"file-id")
        tt = self.get_preview(tree)
        trans_id = tt.trans_id_tree_path("file")
        self.assertEqual((LINES_ONE,), tt._get_parents_texts(trans_id))


class TestOrphan(tests.TestCaseWithTransport):
    def test_no_orphan_for_transform_preview(self):
        tree = self.make_branch_and_tree("tree")
        tt = tree.preview_transform()
        self.addCleanup(tt.finalize)
        self.assertRaises(NotImplementedError, tt.new_orphan, "foo", "bar")

    def _set_orphan_policy(self, wt, policy):
        wt.branch.get_config_stack().set("transform.orphan_policy", policy)

    def _prepare_orphan(self, wt):
        self.build_tree(["dir/", "dir/file", "dir/foo"])
        wt.add(["dir", "dir/file"], ids=[b"dir-id", b"file-id"])
        wt.commit("add dir and file ignoring foo")
        tt = wt.transform()
        self.addCleanup(tt.finalize)
        # dir and bar are deleted
        dir_tid = tt.trans_id_tree_path("dir")
        file_tid = tt.trans_id_tree_path("dir/file")
        orphan_tid = tt.trans_id_tree_path("dir/foo")
        tt.delete_contents(file_tid)
        tt.unversion_file(file_tid)
        tt.delete_contents(dir_tid)
        tt.unversion_file(dir_tid)
        # There should be a conflict because dir still contain foo
        raw_conflicts = tt.find_raw_conflicts()
        self.assertLength(1, raw_conflicts)
        self.assertEqual(("missing parent", "new-1"), raw_conflicts[0])
        return tt, orphan_tid

    def test_new_orphan_created(self):
        wt = self.make_branch_and_tree(".")
        self._set_orphan_policy(wt, "move")
        tt, orphan_tid = self._prepare_orphan(wt)
        warnings = []

        def warning(*args):
            warnings.append(args[0] % args[1:])

        self.overrideAttr(trace, "warning", warning)
        remaining_conflicts = resolve_conflicts(tt)
        self.assertEqual(["dir/foo has been orphaned in brz-orphans"], warnings)
        # Yeah for resolved conflicts !
        self.assertLength(0, remaining_conflicts)
        # We have a new orphan
        self.assertEqual("foo.~1~", tt.final_name(orphan_tid))
        self.assertEqual("brz-orphans", tt.final_name(tt.final_parent(orphan_tid)))

    def test_never_orphan(self):
        wt = self.make_branch_and_tree(".")
        self._set_orphan_policy(wt, "conflict")
        tt, orphan_tid = self._prepare_orphan(wt)
        remaining_conflicts = resolve_conflicts(tt)
        self.assertLength(1, remaining_conflicts)
        self.assertEqual(
            ("deleting parent", "Not deleting", "new-1"), remaining_conflicts.pop()
        )

    def test_orphan_error(self):
        def bogus_orphan(tt, orphan_id, parent_id):
            raise transform.OrphaningError(
                tt.final_name(orphan_id), tt.final_name(parent_id)
            )

        transform.orphaning_registry.register(
            "bogus", bogus_orphan, "Raise an error when orphaning"
        )
        wt = self.make_branch_and_tree(".")
        self._set_orphan_policy(wt, "bogus")
        tt, orphan_tid = self._prepare_orphan(wt)
        remaining_conflicts = resolve_conflicts(tt)
        self.assertLength(1, remaining_conflicts)
        self.assertEqual(
            ("deleting parent", "Not deleting", "new-1"), remaining_conflicts.pop()
        )

    def test_unknown_orphan_policy(self):
        wt = self.make_branch_and_tree(".")
        # Set a fictional policy nobody ever implemented
        self._set_orphan_policy(wt, "donttouchmypreciouuus")
        tt, orphan_tid = self._prepare_orphan(wt)
        warnings = []

        def warning(*args):
            warnings.append(args[0] % args[1:])

        self.overrideAttr(trace, "warning", warning)
        remaining_conflicts = resolve_conflicts(tt)
        # We fallback to the default policy which create a conflict
        self.assertLength(1, remaining_conflicts)
        self.assertEqual(
            ("deleting parent", "Not deleting", "new-1"), remaining_conflicts.pop()
        )
        self.assertLength(1, warnings)
        self.assertStartsWith(warnings[0], 'Value "donttouchmypreciouuus" ')


class TestTransformHooks(tests.TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.wt = self.make_branch_and_tree(".")
        os.chdir("..")

    def transform(self):
        transform = self.wt.transform()
        self.addCleanup(transform.finalize)
        return transform, transform.root

    def test_pre_commit_hooks(self):
        calls = []

        def record_pre_transform(tree, tt):
            calls.append((tree, tt))

        MutableTree.hooks.install_named_hook(
            "pre_transform", record_pre_transform, "Pre transform"
        )
        transform, root = self.transform()
        old_root_id = transform.tree_file_id(root)
        transform.apply()
        self.assertEqual(old_root_id, self.wt.path2id(""))
        self.assertEqual([(self.wt, transform)], calls)

    def test_post_commit_hooks(self):
        calls = []

        def record_post_transform(tree, tt):
            calls.append((tree, tt))

        MutableTree.hooks.install_named_hook(
            "post_transform", record_post_transform, "Post transform"
        )
        transform, root = self.transform()
        old_root_id = transform.tree_file_id(root)
        transform.apply()
        self.assertEqual(old_root_id, self.wt.path2id(""))
        self.assertEqual([(self.wt, transform)], calls)


class TestLinkTree(tests.TestCaseWithTransport):
    def setUp(self):
        tests.TestCaseWithTransport.setUp(self)
        self.requireFeature(HardlinkFeature(self.test_dir))
        self.parent_tree = self.make_branch_and_tree("parent")
        self.parent_tree.lock_write()
        self.addCleanup(self.parent_tree.unlock)
        self.build_tree_contents([("parent/foo", b"bar")])
        self.parent_tree.add("foo")
        self.parent_tree.commit("added foo")
        child_controldir = self.parent_tree.controldir.sprout("child")
        self.child_tree = child_controldir.open_workingtree()

    def hardlinked(self):
        parent_stat = os.lstat(self.parent_tree.abspath("foo"))
        child_stat = os.lstat(self.child_tree.abspath("foo"))
        return parent_stat.st_ino == child_stat.st_ino

    def test_link_fails_if_modified(self):
        """If the file to be linked has modified text, don't link."""
        self.build_tree_contents([("child/foo", b"baz")])
        transform.link_tree(self.child_tree, self.parent_tree)
        self.assertFalse(self.hardlinked())

    def test_link_fails_if_execute_bit_changed(self):
        """If the file to be linked has modified execute bit, don't link."""
        tt = self.child_tree.transform()
        try:
            trans_id = tt.trans_id_tree_path("foo")
            tt.set_executability(True, trans_id)
            tt.apply()
        finally:
            tt.finalize()
        transform.link_tree(self.child_tree, self.parent_tree)
        self.assertFalse(self.hardlinked())

    def test_link_succeeds_if_unmodified(self):
        """If the file to be linked is unmodified, link."""
        transform.link_tree(self.child_tree, self.parent_tree)
        self.assertTrue(self.hardlinked())


class ErrorTests(tests.TestCase):
    def test_transform_rename_failed(self):
        e = TransformRenameFailed("from", "to", "readonly file", 2)
        self.assertEqual("Failed to rename from to to: readonly file", str(e))



# Copyright (C) 2007-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for Launchpad user ID management functions."""

from ... import config
from ...tests import TestCaseInTempDir, TestCaseWithMemoryTransport
from . import account


class LaunchpadAccountTests(TestCaseInTempDir):
    def test_get_lp_login_unconfigured(self):
        # Test that get_lp_login() returns None if no username has
        # been configured.
        my_config = config.MemoryStack(b"")
        self.assertEqual(None, account.get_lp_login(my_config))

    def test_get_lp_login(self):
        # Test that get_lp_login() returns the configured username
        my_config = config.MemoryStack(b"[DEFAULT]\nlaunchpad_username=test-user\n")
        self.assertEqual("test-user", account.get_lp_login(my_config))

    def test_set_lp_login(self):
        # Test that set_lp_login() updates the config file.
        my_config = config.MemoryStack(b"")
        self.assertEqual(None, my_config.get("launchpad_username"))
        account.set_lp_login("test-user", my_config)
        self.assertEqual("test-user", my_config.get("launchpad_username"))

    def test_unknown_launchpad_username(self):
        # Test formatting of UnknownLaunchpadUsername exception
        error = account.UnknownLaunchpadUsername(user="test-user")
        self.assertEqualDiff(
            "The user name test-user is not registered on Launchpad.", str(error)
        )

    def test_no_registered_ssh_keys(self):
        # Test formatting of NoRegisteredSSHKeys exception
        error = account.NoRegisteredSSHKeys(user="test-user")
        self.assertEqualDiff(
            "The user test-user has not registered any "
            "SSH keys with Launchpad.\n"
            "See <https://launchpad.net/people/+me>",
            str(error),
        )

    def test_set_lp_login_updates_authentication_conf(self):
        self.assertIs(None, account._get_auth_user())
        account.set_lp_login("foo")
        self.assertEqual("foo", account._get_auth_user())

    def test_get_lp_login_does_not_update_for_none_user(self):
        account.get_lp_login()
        self.assertIs(None, account._get_auth_user())

    def test_get_lp_login_updates_authentication_conf(self):
        account._set_global_option("foo")
        self.assertIs(None, account._get_auth_user())
        account.get_lp_login()
        auth = config.AuthenticationConfig()
        self.assertEqual("foo", account._get_auth_user(auth))
        self.assertEqual("foo", auth.get_user("ssh", "bazaar.launchpad.net"))
        self.assertEqual("foo", auth.get_user("ssh", "bazaar.staging.launchpad.net"))

    def test_get_lp_login_leaves_existing_credentials(self):
        auth = config.AuthenticationConfig()
        auth.set_credentials("Foo", "bazaar.launchpad.net", "foo", "ssh")
        auth.set_credentials("Bar", "bazaar.staging.launchpad.net", "foo", "ssh")
        account._set_global_option("foo")
        account.get_lp_login()
        auth = config.AuthenticationConfig()
        credentials = auth.get_credentials("ssh", "bazaar.launchpad.net")
        self.assertEqual("Foo", credentials["name"])

    def test_get_lp_login_errors_on_mismatch(self):
        account._set_auth_user("foo")
        account._set_global_option("bar")
        e = self.assertRaises(account.MismatchedUsernames, account.get_lp_login)
        self.assertEqual(
            "breezy.conf and authentication.conf disagree about"
            " launchpad account name.  Please re-run launchpad-login.",
            str(e),
        )


class CheckAccountTests(TestCaseWithMemoryTransport):
    def test_check_lp_login_valid_user(self):
        transport = self.get_transport()
        transport.mkdir("~test-user")
        transport.put_bytes("~test-user/+sshkeys", b"some keys here")
        account.check_lp_login("test-user", transport)

    def test_check_lp_login_no_user(self):
        transport = self.get_transport()
        self.assertRaises(
            account.UnknownLaunchpadUsername,
            account.check_lp_login,
            "test-user",
            transport,
        )

    def test_check_lp_login_no_ssh_keys(self):
        transport = self.get_transport()
        transport.mkdir("~test-user")
        transport.put_bytes("~test-user/+sshkeys", b"")
        self.assertRaises(
            account.NoRegisteredSSHKeys, account.check_lp_login, "test-user", transport
        )



"""Flake8 plugin for reading lazy imports."""

import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))


class LazyImport:
    """A dummy plugin that is present mainly to monkey patch pyflakes."""

    name = "lazy-import"
    version = "0.1"

    def __init__(self, tree):
        self.tree = tree

    def run(self):
        """Do nothing."""
        return []



# Copyright (C) 2006, 2007, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os

from breezy import merge, tests, transform, workingtree


class TestRevert(tests.TestCaseWithTransport):
    """Ensure that revert behaves as expected."""

    def test_revert_merged_dir(self):
        """Reverting a merge that adds a directory deletes the directory."""
        source_tree = self.make_branch_and_tree("source")
        source_tree.commit("empty tree")
        target_tree = source_tree.controldir.sprout("target").open_workingtree()
        self.build_tree(["source/dir/", "source/dir/contents"])
        source_tree.add(["dir", "dir/contents"], ids=[b"dir-id", b"contents-id"])
        source_tree.commit("added dir")
        target_tree.lock_write()
        self.addCleanup(target_tree.unlock)
        merge.merge_inner(
            target_tree.branch,
            source_tree.basis_tree(),
            target_tree.basis_tree(),
            this_tree=target_tree,
        )
        self.assertPathExists("target/dir")
        self.assertPathExists("target/dir/contents")
        target_tree.revert()
        self.assertPathDoesNotExist("target/dir/contents")
        self.assertPathDoesNotExist("target/dir")

    def test_revert_new(self):
        """Only locally-changed new files should be preserved when reverting.

        When a file isn't present in revert's target tree:
        If a file hasn't been committed, revert should unversion it, but not
        delete it.
        If a file has local changes, revert should unversion it, but not
        delete it.
        If a file has no changes from the last commit, revert should delete it.
        If a file has changes due to a merge, revert should delete it.
        """
        tree = self.make_branch_and_tree("tree")
        tree.commit("empty tree")
        merge_target = tree.controldir.sprout("merge_target").open_workingtree()
        self.build_tree(["tree/new_file"])

        # newly-added files should not be deleted
        tree.add("new_file")
        basis_tree = tree.branch.repository.revision_tree(tree.last_revision())
        tree.revert()
        self.assertPathExists("tree/new_file")

        # unchanged files should be deleted
        tree.add("new_file")
        tree.commit("add new_file")
        tree.revert(old_tree=basis_tree)
        self.assertPathDoesNotExist("tree/new_file")

        # files should be deleted if their changes came from merges
        merge_target.merge_from_branch(tree.branch)
        self.assertPathExists("merge_target/new_file")
        merge_target.revert()
        self.assertPathDoesNotExist("merge_target/new_file")

        # files should not be deleted if changed after a merge
        merge_target.merge_from_branch(tree.branch)
        self.assertPathExists("merge_target/new_file")
        self.build_tree_contents([("merge_target/new_file", b"new_contents")])
        merge_target.revert()
        self.assertPathExists("merge_target/new_file")

    def tree_with_executable(self):
        tree = self.make_branch_and_tree("tree")
        tt = tree.transform()
        tt.new_file("newfile", tt.root, [b"helooo!"], b"newfile-id", True)
        tt.apply()
        with tree.lock_write():
            self.assertTrue(tree.is_executable("newfile"))
            tree.commit("added newfile")
        return tree

    def test_preserve_execute(self):
        tree = self.tree_with_executable()
        tt = tree.transform()
        newfile = tt.trans_id_tree_path("newfile")
        tt.delete_contents(newfile)
        tt.create_file([b"Woooorld!"], newfile)
        tt.apply()
        tree = workingtree.WorkingTree.open("tree")
        tree.lock_write()
        self.addCleanup(tree.unlock)
        self.assertTrue(tree.is_executable("newfile"))
        transform.revert(tree, tree.basis_tree(), None, backups=True)
        with tree.get_file("newfile", "rb") as f:
            self.assertEqual(b"helooo!", f.read())
        self.assertTrue(tree.is_executable("newfile"))

    def test_revert_executable(self):
        tree = self.tree_with_executable()
        tt = tree.transform()
        newfile = tt.trans_id_tree_path("newfile")
        tt.set_executability(False, newfile)
        tt.apply()
        tree.lock_write()
        self.addCleanup(tree.unlock)
        transform.revert(tree, tree.basis_tree(), None)
        self.assertTrue(tree.is_executable("newfile"))

    def test_revert_deletes_files_from_revert(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["file"])
        tree.add("file")
        rev1 = tree.commit("added file")
        with tree.lock_read():
            file_sha = tree.get_file_sha1("file")
        os.unlink("file")
        tree.commit("removed file")
        self.assertPathDoesNotExist("file")
        tree.revert(old_tree=tree.branch.repository.revision_tree(rev1))
        self.assertEqual({"file": file_sha}, tree.merge_modified())
        self.assertPathExists("file")
        tree.revert()
        self.assertPathDoesNotExist("file")
        self.assertEqual({}, tree.merge_modified())

    def test_revert_file_in_deleted_dir(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["dir/", "dir/file1", "dir/file2"])
        tree.add(
            ["dir", "dir/file1", "dir/file2"], ids=[b"dir-id", b"file1-id", b"file2-id"]
        )
        tree.commit("Added files")
        os.unlink("dir/file1")
        os.unlink("dir/file2")
        os.rmdir("dir")
        tree.remove(["dir/", "dir/file1", "dir/file2"])
        tree.revert(["dir/file1"])
        self.assertPathExists("dir/file1")
        self.assertPathDoesNotExist("dir/file2")
        self.assertEqual(b"dir-id", tree.path2id("dir"))

    def test_revert_root_id_change(self):
        tree = self.make_branch_and_tree(".")
        tree.set_root_id(b"initial-root-id")
        self.build_tree(["file1"])
        tree.add(["file1"])
        tree.commit("first")
        tree.set_root_id(b"temp-root-id")
        self.assertEqual(b"temp-root-id", tree.path2id(""))
        tree.revert()
        self.assertEqual(b"initial-root-id", tree.path2id(""))



# Copyright (C) 2007-2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for commands related to tags."""

from breezy import branch as _mod_branch
from breezy import errors, tag
from breezy.branch import Branch
from breezy.bzr import branch as bzrbranch
from breezy.bzr import bzrdir
from breezy.tests import TestCaseWithTransport, script
from breezy.workingtree import WorkingTree


class TestTagging(TestCaseWithTransport):
    def test_tag_command_help(self):
        out, err = self.run_bzr("help tag")
        self.assertContainsRe(out, "Create, remove or modify a tag")

    def test_cannot_tag_range(self):
        out, err = self.run_bzr("tag -r1..10 name", retcode=3)
        self.assertContainsRe(err, "Tags can only be placed on a single revision")

    def test_no_tag_name(self):
        out, err = self.run_bzr("tag -d branch", retcode=3)
        self.assertContainsRe(err, "Please specify a tag name.")

    def test_automatic_tag_name(self):
        def get_tag_name(branch, revid):
            return "mytag"

        Branch.hooks.install_named_hook(
            "automatic_tag_name", get_tag_name, "get tag name"
        )
        out, err = self.run_bzr("tag -d branch")
        self.assertContainsRe(err, "Created tag mytag.")

    def test_tag_current_rev(self):
        t = self.make_branch_and_tree("branch")
        t.commit(allow_pointless=True, message="initial commit", rev_id=b"first-revid")
        # make a tag through the command line
        out, err = self.run_bzr("tag -d branch NEWTAG")
        self.assertContainsRe(err, "Created tag NEWTAG.")
        # tag should be observable through the api
        self.assertEqual(t.branch.tags.get_tag_dict(), {"NEWTAG": b"first-revid"})
        # can also create tags using -r
        self.run_bzr("tag -d branch tag2 -r1")
        self.assertEqual(t.branch.tags.lookup_tag("tag2"), b"first-revid")
        # regression test: make sure a unicode revision from the user
        # gets turned into a str object properly. The use of a unicode
        # object for the revid is intentional.
        self.run_bzr(["tag", "-d", "branch", "tag3", "-rrevid:first-revid"])
        self.assertEqual(t.branch.tags.lookup_tag("tag3"), b"first-revid")
        # can also delete an existing tag
        out, err = self.run_bzr("tag --delete -d branch tag2")
        # cannot replace an existing tag normally
        out, err = self.run_bzr("tag -d branch NEWTAG -r0", retcode=3)
        self.assertContainsRe(err, "Tag NEWTAG already exists\\.")
        # ... but can if you use --force
        out, err = self.run_bzr("tag -d branch NEWTAG --force -r0")
        self.assertEqual("Updated tag NEWTAG.\n", err)

    def test_tag_same_revision(self):
        t = self.make_branch_and_tree("branch")
        t.commit(allow_pointless=True, message="initial commit", rev_id=b"first-revid")
        t.commit(allow_pointless=True, message="second commit", rev_id=b"second-revid")
        out, err = self.run_bzr("tag -rrevid:first-revid -d branch NEWTAG")
        out, err = self.run_bzr("tag -rrevid:first-revid -d branch NEWTAG")
        self.assertContainsRe(err, "Tag NEWTAG already exists for that revision\\.")
        out, err = self.run_bzr("tag -rrevid:second-revid -d branch NEWTAG", retcode=3)
        self.assertContainsRe(err, "Tag NEWTAG already exists\\.")

    def test_tag_delete_requires_name(self):
        out, err = self.run_bzr("tag -d branch", retcode=3)
        self.assertContainsRe(err, "Please specify a tag name\\.")

    def test_branch_push_pull_merge_copies_tags(self):
        t = self.make_branch_and_tree("branch1")
        t.commit(allow_pointless=True, message="initial commit", rev_id=b"first-revid")
        b1 = t.branch
        b1.tags.set_tag("tag1", b"first-revid")
        # branching copies the tag across
        self.run_bzr("branch branch1 branch2")
        b2 = Branch.open("branch2")
        self.assertEqual(b2.tags.lookup_tag("tag1"), b"first-revid")
        # make a new tag and pull it
        b1.tags.set_tag("tag2", b"twa")
        self.run_bzr("pull -d branch2 branch1")
        self.assertEqual(b2.tags.lookup_tag("tag2"), b"twa")
        # make a new tag and push it
        b1.tags.set_tag("tag3", b"san")
        self.run_bzr("push -d branch1 branch2")
        self.assertEqual(b2.tags.lookup_tag("tag3"), b"san")
        # make a new tag and merge it
        t.commit(allow_pointless=True, message="second commit", rev_id=b"second-revid")
        t2 = WorkingTree.open("branch2")
        t2.commit(allow_pointless=True, message="commit in second")
        b1.tags.set_tag("tag4", b"second-revid")
        self.run_bzr("merge -d branch2 branch1")
        self.assertEqual(b2.tags.lookup_tag("tag4"), b"second-revid")
        # pushing to a new location copies the tag across
        self.run_bzr("push -d branch1 branch3")
        b3 = Branch.open("branch3")
        self.assertEqual(b3.tags.lookup_tag("tag1"), b"first-revid")

    def make_master_and_checkout(self):
        builder = self.make_branch_builder("master")
        builder.build_commit(message="Initial commit.", rev_id=b"rev-1")
        master = builder.get_branch()
        child = master.create_checkout(self.get_url("child"))
        return master, child

    def make_fork(self, branch):
        fork = branch.create_clone_on_transport(self.get_transport("fork"))
        self.addCleanup(fork.lock_write().unlock)
        with fork.basis_tree().preview_transform() as tt:
            tt.commit(fork, message="Commit in fork.", revision_id=b"fork-0")
        with fork.basis_tree().preview_transform() as tt:
            tt.commit(fork, message="Commit in fork.", revision_id=b"fork-1")
        return fork

    def test_merge_without_commit_does_not_propagate_tags_to_master(self):
        """'brz merge' alone does not propagate tags to a master branch.

        (If the user runs 'brz commit', then that is when the tags from the
        merge are propagated.)
        """
        master, child = self.make_master_and_checkout()
        fork = self.make_fork(master)
        fork.tags.set_tag("new-tag", fork.last_revision())
        self.run_bzr(["merge", "../fork"], working_dir="child")
        self.assertEqual({}, master.tags.get_tag_dict())

    def test_commit_in_heavyweight_checkout_copies_tags_to_master(self):
        master, child = self.make_master_and_checkout()
        fork = self.make_fork(master)
        fork.tags.set_tag("new-tag", fork.last_revision())
        fork.tags.set_tag("non-ancestry-tag", b"fork-0")
        fork.tags.set_tag("absent-tag", b"absent-rev")
        script.run_script(
            self,
            """
            $ cd child
            $ brz merge ../fork
            $ brz commit -m "Merge fork."
            2>Committing to: .../master/
            2>Committed revision 2.
            """,
            null_output_matches_anything=True,
        )
        # Merge copied the tag to child and commit propagated it to master
        expected_tag_dict = {
            "new-tag": fork.last_revision(),
            "non-ancestry-tag": b"fork-0",
            "absent-tag": b"absent-rev",
        }
        self.assertEqual(expected_tag_dict, child.branch.tags.get_tag_dict())
        self.assertEqual(expected_tag_dict, master.tags.get_tag_dict())
        # Revisions not in ancestry but named in tags are present
        child.branch.repository.get_revision(b"fork-0")
        master.repository.get_revision(b"fork-0")

    def test_commit_in_heavyweight_checkout_reports_tag_conflict(self):
        master, child = self.make_master_and_checkout()
        fork = self.make_fork(master)
        fork.tags.set_tag("new-tag", fork.last_revision())
        master_r1 = master.last_revision()
        master.tags.set_tag("new-tag", master_r1)
        script.run_script(
            self,
            """
            $ cd child
            $ brz merge ../fork
            $ brz commit -m "Merge fork."
            2>Committing to: .../master/
            2>Conflicting tags in bound branch:
            2>    new-tag
            2>Committed revision 2.
            """,
            null_output_matches_anything=True,
        )
        # Merge copied the tag to child.  master's conflicting tag is unchanged.
        self.assertEqual(
            {"new-tag": fork.last_revision()}, child.branch.tags.get_tag_dict()
        )
        self.assertEqual({"new-tag": master_r1}, master.tags.get_tag_dict())

    def test_list_tags(self):
        tree1 = self.make_branch_and_tree("branch1")
        tree1.commit(
            allow_pointless=True, message="revision 1", rev_id=b"revid-1", timestamp=10
        )
        tree1.commit(
            allow_pointless=True, message="revision 2", rev_id=b"revid-2", timestamp=15
        )

        b1 = tree1.branch
        # note how the tag for revid-1 sorts after the one for revid-2
        b1.tags.set_tag("tag1\u30d0", b"revid-2")
        # not present in repository
        b1.tags.set_tag("tag10\u30d0", b"missing")
        b1.tags.set_tag("tag2\u30d0", b"revid-1")

        # natural order
        out, err = self.run_bzr_raw("tags -d branch1", encoding="utf-8")
        self.assertEqual(err, b"")
        self.assertContainsRe(
            out,
            ("^tag1\u30d0  *2\ntag2\u30d0  *1\n" + "tag10\u30d0 *\\?\n").encode(
                "utf-8"
            ),
        )

        # lexicographical order
        out, err = self.run_bzr_raw("tags --sort=alpha -d branch1", encoding="utf-8")
        self.assertEqual(err, b"")
        self.assertContainsRe(
            out,
            ("^tag10\u30d0  *\\?\ntag1\u30d0  *2\n" + "tag2\u30d0 *1\n").encode(
                "utf-8"
            ),
        )

        out, err = self.run_bzr_raw(
            "tags --sort=alpha --show-ids -d branch1", encoding="utf-8"
        )
        self.assertEqual(err, b"")
        self.assertContainsRe(
            out,
            (
                "^tag10\u30d0  *missing\n"
                + "tag1\u30d0  *revid-2\ntag2\u30d0 *revid-1\n"
            ).encode("utf-8"),
        )

        # chronological order
        out, err = self.run_bzr_raw("tags --sort=time -d branch1", encoding="utf-8")
        self.assertEqual(err, b"")
        self.assertContainsRe(
            out,
            ("^tag2\u30d0  *1\ntag1\u30d0  *2\n" + "tag10\u30d0 *\\?\n").encode(
                "utf-8"
            ),
        )

        out, err = self.run_bzr_raw(
            "tags --sort=time --show-ids -d branch1", encoding="utf-8"
        )
        self.assertEqual(err, b"")
        self.assertContainsRe(
            out,
            (
                "^tag2\u30d0  *revid-1\n"
                + "tag1\u30d0  *revid-2\ntag10\u30d0 *missing\n"
            ).encode("utf-8"),
        )

        # now test dotted revnos
        tree2 = tree1.controldir.sprout("branch2").open_workingtree()
        tree1.commit(
            allow_pointless=True, message="revision 3 in branch1", rev_id=b"revid-3a"
        )
        tree2.commit(
            allow_pointless=True, message="revision 3 in branch2", rev_id=b"revid-3b"
        )

        b2 = tree2.branch
        b2.tags.set_tag("tagD", b"revid-3b")
        self.run_bzr("merge -d branch1 branch2")
        tree1.commit("merge", rev_id=b"revid-4")

        out, err = self.run_bzr("tags -d branch1", encoding="utf-8")
        self.assertEqual(err, "")
        self.assertContainsRe(out, r"tagD  *2\.1\.1\n")
        out, err = self.run_bzr("tags -d branch2", encoding="utf-8")
        self.assertEqual(err, "")
        self.assertContainsRe(out, r"tagD  *3\n")

    def test_list_tags_dotted_revnos_unsupported(self):
        class TrimmedBranch(bzrbranch.BzrBranch6):
            def revision_id_to_dotted_revno(self, revid):
                raise errors.UnsupportedOperation(
                    self.revision_id_to_dotted_revno, self
                )

        class TrimmedBranchFormat(bzrbranch.BzrBranchFormat6):
            def _branch_class(self):
                return TrimmedBranch

            @classmethod
            def get_format_string(cls):
                return b"Trimmed Branch"

        _mod_branch.format_registry.register(TrimmedBranchFormat())
        self.addCleanup(_mod_branch.format_registry.remove, TrimmedBranchFormat())

        dir_format = bzrdir.BzrDirMetaFormat1()
        dir_format._branch_format = TrimmedBranchFormat()

        tree = self.make_branch_and_tree("branch", format=dir_format)
        self.assertFileEqual("Trimmed Branch", "branch/.bzr/branch/format")
        rev1 = tree.commit("rev1")
        tree.branch.tags.set_tag("mytag", rev1)
        out, err = self.run_bzr("tags -d branch", encoding="utf-8")
        self.assertEqual(out, "mytag                ?\n")

    def test_list_tags_revision_filtering(self):
        tree1 = self.make_branch_and_tree(".")
        tree1.commit(allow_pointless=True, message="revision 1", rev_id=b"revid-1")
        tree1.commit(allow_pointless=True, message="revision 2", rev_id=b"revid-2")
        tree1.commit(allow_pointless=True, message="revision 3", rev_id=b"revid-3")
        tree1.commit(allow_pointless=True, message="revision 4", rev_id=b"revid-4")
        b1 = tree1.branch
        b1.tags.set_tag("tag 1", b"revid-1")
        b1.tags.set_tag("tag 2", b"revid-2")
        b1.tags.set_tag("tag 3", b"revid-3")
        b1.tags.set_tag("tag 4", b"revid-4")
        self._check_tag_filter("", (1, 2, 3, 4))
        self._check_tag_filter("-r ..", (1, 2, 3, 4))
        self._check_tag_filter("-r ..2", (1, 2))
        self._check_tag_filter("-r 2..", (2, 3, 4))
        self._check_tag_filter("-r 2..3", (2, 3))
        self._check_tag_filter("-r 3..2", ())
        self.run_bzr_error(
            args="tags -r 123",
            error_regexes=[
                "brz: ERROR: Requested revision: '123' does not exist in branch:"
            ],
        )
        self.run_bzr_error(
            args="tags -r ..123",
            error_regexes=[
                "brz: ERROR: Requested revision: '123' does not exist in branch:"
            ],
        )
        self.run_bzr_error(
            args="tags -r 123.123",
            error_regexes=[
                "brz: ERROR: Requested revision: '123.123' does not exist in branch:"
            ],
        )

    def test_sort_tags_custom(self):
        def sort_by_dots(branch, tags):
            def sort_key(tag_and_revid):
                return tag_and_revid[0].count(".")

            tags.sort(key=sort_key)

        # Register a custom sort method
        tag.tag_sort_methods.register("dots", sort_by_dots, "Sort by dots.")
        self.addCleanup(tag.tag_sort_methods.remove, "dots")

        tree1 = self.make_branch_and_tree("branch1")
        tree1.commit(
            allow_pointless=True, message="revision 1", rev_id=b"revid-1", timestamp=10
        )
        tree1.commit(
            allow_pointless=True, message="revision 2", rev_id=b"revid-2", timestamp=15
        )

        b1 = tree1.branch

        b1.tags.set_tag("tag..", b"revid-2")
        b1.tags.set_tag("tag....", b"missing")  # not present in repository
        b1.tags.set_tag("tag.", b"revid-1")
        b1.tags.set_tag("tag...", b"revid-1")
        b1.tags.set_tag("tag....", b"revid-1")

        # sorted by number of dots
        out, err = self.run_bzr("tags --sort=dots -d branch1")
        self.assertEqual(err, "")
        self.assertEqual(
            [
                "tag.                 1",
                "tag..                2",
                "tag...               1",
                "tag....              1",
            ],
            out.splitlines(),
        )

    def _check_tag_filter(self, argstr, expected_revnos):
        # upper bound of laziness
        out, err = self.run_bzr("tags " + argstr)
        self.assertEqual(err, "")
        self.assertContainsRe(
            out,
            "^"
            + "".join(
                ["tag {} +{}\n".format(revno, revno) for revno in expected_revnos]
            )
            + "$",
        )

    def test_conflicting_tags(self):
        # setup two empty branches with different tags
        t1 = self.make_branch_and_tree("one")
        t2 = self.make_branch_and_tree("two")
        b1 = t1.branch
        b2 = t2.branch
        tagname = "\u30d0zaar"
        b1.tags.set_tag(tagname, b"revid1")
        b2.tags.set_tag(tagname, b"revid2")
        # push should give a warning about the tags
        out, err = self.run_bzr("push -d one two", encoding="utf-8")
        self.assertContainsRe(out, "Conflicting tags:\n.*" + tagname)
        # pull should give a warning about the tags
        out, err = self.run_bzr("pull -d one two", encoding="utf-8", retcode=1)
        self.assertContainsRe(out, "Conflicting tags:\n.*" + tagname)
        # merge should give a warning about the tags -- not implemented yet
        ## out, err = self.run_bzr('merge -d one two', encoding='utf-8')
        # self.assertContainsRe(out,
        # 'Conflicting tags:\n.*' + tagname.encode('utf-8'))

    def test_tag_quiet(self):
        self.make_branch_and_tree("")
        out, err = self.run_bzr("tag --quiet test1")
        self.assertEqual("", out)
        self.assertEqual("", err)

    def test_tag_delete_quiet(self):
        self.make_branch_and_tree("")
        self.run_bzr("tag test1")
        out, err = self.run_bzr("tag --delete --quiet test1")
        self.assertEqual("", out)
        self.assertEqual("", err)

    def test_tags_with_mainline_ghosts(self):
        tree = self.make_branch_and_tree("tree1")
        tree.set_parent_ids([b"spooky"], allow_leftmost_as_ghost=True)
        tree.add("")
        tree.commit("msg1", rev_id=b"rev1")
        tree.commit("msg2", rev_id=b"rev2")
        tree.branch.tags.set_tag("unknown", b"out-of-mainline")
        tree.branch.tags.set_tag("ghost", b"spooky")
        tree.branch.tags.set_tag("tag1", b"rev1")
        tree.branch.tags.set_tag("tag2", b"rev2")

        out, err = self.run_bzr("tags -d tree1", encoding="utf-8")
        self.assertEqual(
            out,
            "ghost                ?\n"
            "tag1                 1\n"
            "tag2                 2\n"
            "unknown              ?\n",
        )
        self.assertEqual("", err)

    def test_tag_unsupported(self):
        self.make_branch_and_tree("tree", format="dirstate")
        out, err = self.run_bzr("tag -d tree blah", retcode=3)
        self.assertEqual(out, "")
        self.assertContainsRe(
            err,
            "brz: ERROR: Tags not supported by BzrBranch5\\(.*\\/tree\\/\\); "
            "you may be able to use 'brz upgrade file:\\/\\/.*\\/tree\\/'.",
        )



# Copyright (C) 2007-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from breezy.tests import TestCaseWithTransport


class TestCatRevision(TestCaseWithTransport):
    def test_cat_unicode_revision(self):
        tree = self.make_branch_and_tree(".")
        tree.commit("This revision", rev_id=b"abcd")
        output, errors = self.run_bzr(["cat-revision", "abcd"])
        self.assertContainsRe(output, "This revision")
        self.assertEqual("", errors)

    def test_cat_revision(self):
        """Test brz cat-revision."""
        wt = self.make_branch_and_tree(".")
        r = wt.branch.repository

        wt.commit("Commit one", rev_id=b"a@r-0-1")
        wt.commit("Commit two", rev_id=b"a@r-0-2")
        wt.commit("Commit three", rev_id=b"a@r-0-3")

        with r.lock_read():
            revs = {}
            for i in (1, 2, 3):
                revid = b"a@r-0-%d" % i
                stream = r.revisions.get_record_stream([(revid,)], "unordered", False)
                revs[i] = next(stream).get_bytes_as("fulltext")

        for i in [1, 2, 3]:
            self.assertEqual(
                revs[i],
                self.run_bzr("cat-revision -r revid:a@r-0-%d" % i)[0].encode("utf-8"),
            )
            self.assertEqual(
                revs[i], self.run_bzr("cat-revision a@r-0-%d" % i)[0].encode("utf-8")
            )
            self.assertEqual(
                revs[i], self.run_bzr("cat-revision -r %d" % i)[0].encode("utf-8")
            )

    def test_cat_no_such_revid(self):
        self.make_branch_and_tree(".")
        err = self.run_bzr("cat-revision abcd", retcode=3)[1]
        self.assertContainsRe(err, "The repository .* contains no revision abcd.")

    def test_cat_revision_directory(self):
        """Test --directory option."""
        tree = self.make_branch_and_tree("a")
        tree.commit("This revision", rev_id=b"abcd")
        output, errors = self.run_bzr(["cat-revision", "-d", "a", "abcd"])
        self.assertContainsRe(output, "This revision")
        self.assertEqual("", errors)

    def test_cat_tree_less_branch(self):
        tree = self.make_branch_and_tree(".")
        tree.commit("This revision", rev_id=b"abcd")
        tree.controldir.destroy_workingtree()
        output, errors = self.run_bzr(["cat-revision", "-d", "a", "abcd"])
        self.assertContainsRe(output, "This revision")
        self.assertEqual("", errors)



# Copyright (C) 2007, 2009, 2011, 2014, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from email.header import decode_header

from .. import __version__ as _breezy_version
from .. import tests
from ..email_message import EmailMessage
from ..errors import BzrBadParameterNotUnicode
from ..smtp_connection import SMTPConnection

EMPTY_MESSAGE = """\
From: from@from.com
Subject: subject
To: to@to.com
User-Agent: Bazaar ({})

""".format(_breezy_version)

_SIMPLE_MESSAGE = """\
MIME-Version: 1.0
Content-Type: text/plain; charset="%s"
Content-Transfer-Encoding: %s
From: from@from.com
Subject: subject
To: to@to.com
User-Agent: Bazaar ({})

%s""".format(_breezy_version)

SIMPLE_MESSAGE_ASCII = _SIMPLE_MESSAGE % ("us-ascii", "7bit", "body")
SIMPLE_MESSAGE_UTF8 = _SIMPLE_MESSAGE % ("utf-8", "base64", "YsOzZHk=\n")
SIMPLE_MESSAGE_8BIT = _SIMPLE_MESSAGE % ("8-bit", "base64", "YvRkeQ==\n")


BOUNDARY = "=====123456=="

_MULTIPART_HEAD = """\
Content-Type: multipart/mixed; boundary="{boundary}"
MIME-Version: 1.0
From: from@from.com
Subject: subject
To: to@to.com
User-Agent: Bazaar ({version})

--{boundary}
MIME-Version: 1.0
Content-Type: text/plain; charset="us-ascii"
Content-Transfer-Encoding: 7bit
Content-Disposition: inline

body
""".format(version=_breezy_version, boundary=BOUNDARY)


def simple_multipart_message():
    msg = _MULTIPART_HEAD + "--{}--\n".format(BOUNDARY)
    return msg


def complex_multipart_message(typ):
    msg = (
        _MULTIPART_HEAD
        + """\
--{boundary}
MIME-Version: 1.0
Content-Type: text/%s; charset="us-ascii"; name="lines.txt"
Content-Transfer-Encoding: 7bit
Content-Disposition: inline

a
b
c
d
e

--{boundary}--
""".format(boundary=BOUNDARY)
    )
    return msg % (typ,)


class TestEmailMessage(tests.TestCase):
    def test_empty_message(self):
        msg = EmailMessage("from@from.com", "to@to.com", "subject")
        self.assertEqualDiff(EMPTY_MESSAGE, msg.as_string())

    def test_simple_message(self):
        pairs = {
            b"body": SIMPLE_MESSAGE_ASCII,
            "b\xf3dy": SIMPLE_MESSAGE_UTF8,
            b"b\xc3\xb3dy": SIMPLE_MESSAGE_UTF8,
            b"b\xf4dy": SIMPLE_MESSAGE_8BIT,
        }
        for body, expected in pairs.items():
            msg = EmailMessage("from@from.com", "to@to.com", "subject", body)
            self.assertEqualDiff(expected, msg.as_string())

    def test_multipart_message_simple(self):
        msg = EmailMessage("from@from.com", "to@to.com", "subject")
        msg.add_inline_attachment("body")
        self.assertEqualDiff(simple_multipart_message(), msg.as_string(BOUNDARY))

    def test_multipart_message_complex(self):
        msg = EmailMessage("from@from.com", "to@to.com", "subject", "body")
        msg.add_inline_attachment("a\nb\nc\nd\ne\n", "lines.txt", "x-subtype")
        self.assertEqualDiff(
            complex_multipart_message("x-subtype"), msg.as_string(BOUNDARY)
        )

    def test_headers_accept_unicode_and_utf8(self):
        for user in [
            "Pepe P\xe9rez <pperez@ejemplo.com>",
            "Pepe P\xc3\xa9red <pperez@ejemplo.com>",
        ]:
            msg = EmailMessage(user, user, user)  # no exception raised

            for header in ["From", "To", "Subject"]:
                value = msg[header]
                value.encode("ascii")  # no UnicodeDecodeError

    def test_headers_reject_8bit(self):
        for i in range(3):  # from_address, to_address, subject
            x = [b'"J. Random Developer" <jrandom@example.com>'] * 3
            x[i] = b"Pepe P\xe9rez <pperez@ejemplo.com>"
            self.assertRaises(BzrBadParameterNotUnicode, EmailMessage, *x)

    def test_multiple_destinations(self):
        to_addresses = ["to1@to.com", "to2@to.com", "to3@to.com"]
        msg = EmailMessage("from@from.com", to_addresses, "subject")
        self.assertContainsRe(
            msg.as_string(), "To: " + ", ".join(to_addresses)
        )  # re.M can't be passed, so no ^$

    def test_retrieving_headers(self):
        msg = EmailMessage("from@from.com", "to@to.com", "subject")
        for header, value in [
            ("From", "from@from.com"),
            ("To", "to@to.com"),
            ("Subject", "subject"),
        ]:
            self.assertEqual(value, msg.get(header))
            self.assertEqual(value, msg[header])
        self.assertEqual(None, msg.get("Does-Not-Exist"))
        self.assertEqual(None, msg["Does-Not-Exist"])
        self.assertEqual("None", msg.get("Does-Not-Exist", "None"))

    def test_setting_headers(self):
        msg = EmailMessage("from@from.com", "to@to.com", "subject")
        msg["To"] = "to2@to.com"
        msg["Cc"] = "cc@cc.com"
        self.assertEqual("to2@to.com", msg["To"])
        self.assertEqual("cc@cc.com", msg["Cc"])

    def test_address_to_encoded_header(self):
        def decode(s):
            """Convert a RFC2047-encoded string to a unicode string."""
            return "".join(
                [
                    chunk.decode(encoding or "ascii")
                    for chunk, encoding in decode_header(s)
                ]
            )

        address = "jrandom@example.com"
        encoded = EmailMessage.address_to_encoded_header(address)
        self.assertEqual(address, encoded)

        address = "J Random Developer <jrandom@example.com>"
        encoded = EmailMessage.address_to_encoded_header(address)
        self.assertEqual(address, encoded)

        address = '"J. Random Developer" <jrandom@example.com>'
        encoded = EmailMessage.address_to_encoded_header(address)
        self.assertEqual(address, encoded)

        address = "Pepe P\xe9rez <pperez@ejemplo.com>"  # unicode ok
        encoded = EmailMessage.address_to_encoded_header(address)
        # addr must be unencoded
        self.assertTrue("pperez@ejemplo.com" in encoded)
        self.assertEqual(address, decode(encoded))

        address = b"Pepe P\xe9rez <pperez@ejemplo.com>"  # ISO-8859-1 not ok
        self.assertRaises(
            BzrBadParameterNotUnicode, EmailMessage.address_to_encoded_header, address
        )

    def test_string_with_encoding(self):
        pairs = {
            "Pepe": (b"Pepe", "ascii"),
            "P\xe9rez": (b"P\xc3\xa9rez", "utf-8"),
            b"P\xc3\xa9rez": (b"P\xc3\xa9rez", "utf-8"),
            b"P\xe8rez": (b"P\xe8rez", "8-bit"),
        }
        for string_, pair in pairs.items():
            self.assertEqual(pair, EmailMessage.string_with_encoding(string_))


class TestSend(tests.TestCase):
    def setUp(self):
        super().setUp()
        self.messages = []

        def send_as_append(_self, msg):
            self.messages.append(msg.as_string(BOUNDARY))

        self.overrideAttr(SMTPConnection, "send_email", send_as_append)

    def send_email(
        self, attachment=None, attachment_filename=None, attachment_mime_subtype="plain"
    ):
        class FakeConfig:
            def get(self, option):
                return None

        EmailMessage.send(
            FakeConfig(),
            "from@from.com",
            "to@to.com",
            "subject",
            "body",
            attachment=attachment,
            attachment_filename=attachment_filename,
            attachment_mime_subtype=attachment_mime_subtype,
        )

    def assertMessage(self, expected):
        self.assertLength(1, self.messages)
        self.assertEqualDiff(expected, self.messages[0])

    def test_send_plain(self):
        self.send_email("a\nb\nc\nd\ne\n", "lines.txt")
        self.assertMessage(complex_multipart_message("plain"))

    def test_send_patch(self):
        self.send_email("a\nb\nc\nd\ne\n", "lines.txt", "x-patch")
        self.assertMessage(complex_multipart_message("x-patch"))

    def test_send_simple(self):
        self.send_email()
        self.assertMessage(SIMPLE_MESSAGE_ASCII)



# Copyright (C) 2006-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the formatting and construction of errors."""

import inspect
import re

from .. import controldir, errors, osutils, tests, urlutils


class TestErrors(tests.TestCase):
    def test_no_arg_named_message(self):
        """Ensure the __init__ and _fmt in errors do not have "message" arg.

        This test fails if __init__ or _fmt in errors has an argument
        named "message" as this can cause errors in some Python versions.
        Python 2.5 uses a slot for StandardError.message.
        See bug #603461
        """
        fmt_pattern = re.compile("%\\(message\\)[sir]")
        for c in errors.BzrError.__subclasses__():
            init = getattr(c, "__init__", None)
            fmt = getattr(c, "_fmt", None)
            if init:
                args = inspect.getfullargspec(init)[0]
                self.assertFalse(
                    "message" in args,
                    (
                        'Argument name "message" not allowed for '
                        '"errors.{}.__init__"'.format(c.__name__)
                    ),
                )
            if fmt and fmt_pattern.search(fmt):
                self.assertFalse(
                    True,
                    ('"message" not allowed in "errors.{}._fmt"'.format(c.__name__)),
                )

    def test_duplicate_help_prefix(self):
        error = errors.DuplicateHelpPrefix("foo")
        self.assertEqualDiff(
            "The prefix foo is in the help search path twice.", str(error)
        )

    def test_ghost_revisions_have_no_revno(self):
        error = errors.GhostRevisionsHaveNoRevno("target", "ghost_rev")
        self.assertEqualDiff(
            "Could not determine revno for {target} because"
            " its ancestry shows a ghost at {ghost_rev}",
            str(error),
        )

    def test_incompatibleVersion(self):
        error = errors.IncompatibleVersion("module", [(4, 5, 6), (7, 8, 9)], (1, 2, 3))
        self.assertEqualDiff(
            "API module is not compatible; one of versions "
            "[(4, 5, 6), (7, 8, 9)] is required, but current version is "
            "(1, 2, 3).",
            str(error),
        )

    def test_inconsistent_delta(self):
        error = errors.InconsistentDelta("path", "file-id", "reason for foo")
        self.assertEqualDiff(
            "An inconsistent delta was supplied involving 'path', 'file-id'\n"
            "reason: reason for foo",
            str(error),
        )

    def test_inconsistent_delta_delta(self):
        error = errors.InconsistentDeltaDelta([], "reason")
        self.assertEqualDiff(
            "An inconsistent delta was supplied: []\nreason: reason", str(error)
        )

    def test_in_process_transport(self):
        error = errors.InProcessTransport("fpp")
        self.assertEqualDiff(
            "The transport 'fpp' is only accessible within this process.", str(error)
        )

    def test_invalid_http_range(self):
        error = errors.InvalidHttpRange(
            "path", "Content-Range: potatoes 0-00/o0oo0", "bad range"
        )
        self.assertEqual(
            "Invalid http range"
            " 'Content-Range: potatoes 0-00/o0oo0'"
            " for path: bad range",
            str(error),
        )

    def test_invalid_range(self):
        error = errors.InvalidRange("path", 12, "bad range")
        self.assertEqual("Invalid range access in path at 12: bad range", str(error))

    def test_jail_break(self):
        error = errors.JailBreak("some url")
        self.assertEqualDiff(
            "An attempt to access a url outside the server jail was made: 'some url'.",
            str(error),
        )

    def test_lock_active(self):
        error = errors.LockActive("lock description")
        self.assertEqualDiff(
            "The lock for 'lock description' is in use and cannot be broken.",
            str(error),
        )

    def test_lock_corrupt(self):
        error = errors.LockCorrupt("corruption info")
        self.assertEqualDiff(
            "Lock is apparently held, but corrupted: "
            "corruption info\n"
            "Use 'brz break-lock' to clear it",
            str(error),
        )

    def test_medium_not_connected(self):
        error = errors.MediumNotConnected("a medium")
        self.assertEqualDiff("The medium 'a medium' is not connected.", str(error))

    def test_no_smart_medium(self):
        error = errors.NoSmartMedium("a transport")
        self.assertEqualDiff(
            "The transport 'a transport' cannot tunnel the smart protocol.",
            str(error),
        )

    def test_no_such_id(self):
        error = errors.NoSuchId("atree", "anid")
        self.assertEqualDiff(
            'The file id "anid" is not present in the tree atree.', str(error)
        )

    def test_no_such_revision_in_tree(self):
        error = errors.NoSuchRevisionInTree("atree", "anid")
        self.assertEqualDiff(
            "The revision id {anid} is not present in the tree atree.", str(error)
        )
        self.assertIsInstance(error, errors.NoSuchRevision)

    def test_not_stacked(self):
        error = errors.NotStacked("a branch")
        self.assertEqualDiff("The branch 'a branch' is not stacked.", str(error))

    def test_not_write_locked(self):
        error = errors.NotWriteLocked("a thing to repr")
        self.assertEqualDiff(
            "'a thing to repr' is not write locked but needs to be.", str(error)
        )

    def test_lock_failed(self):
        error = errors.LockFailed("http://canonical.com/", "readonly transport")
        self.assertEqualDiff(
            "Cannot lock http://canonical.com/: readonly transport", str(error)
        )
        self.assertFalse(error.internal_error)

    def test_unstackable_location(self):
        error = errors.UnstackableLocationError("foo", "bar")
        self.assertEqualDiff("The branch 'foo' cannot be stacked on 'bar'.", str(error))

    def test_unstackable_repository_format(self):
        format = "foo"
        url = "/foo"
        error = errors.UnstackableRepositoryFormat(format, url)
        self.assertEqualDiff(
            "The repository '/foo'(foo) is not a stackable format. "
            "You will need to upgrade the repository to permit branch stacking.",
            str(error),
        )

    def test_up_to_date(self):
        error = errors.UpToDateFormat("someformat")
        self.assertEqualDiff(
            "The branch format someformat is already at the most recent format.",
            str(error),
        )

    def test_read_error(self):
        # a unicode path to check that %r is being used.
        path = "a path"
        error = errors.ReadError(path)
        self.assertContainsRe(str(error), "^Error reading from u?'a path'.$")

    def test_bzrerror_from_literal_string(self):
        # Some code constructs BzrError from a literal string, in which case
        # no further formatting is done.  (I'm not sure raising the base class
        # is a great idea, but if the exception is not intended to be caught
        # perhaps no more is needed.)
        try:
            raise errors.BzrError("this is my errors; %d is not expanded")
        except errors.BzrError as e:
            self.assertEqual("this is my errors; %d is not expanded", str(e))

    def test_reading_completed(self):
        error = errors.ReadingCompleted("a request")
        self.assertEqualDiff(
            "The MediumRequest 'a request' has already had "
            "finish_reading called upon it - the request has been completed and"
            " no more data may be read.",
            str(error),
        )

    def test_writing_completed(self):
        error = errors.WritingCompleted("a request")
        self.assertEqualDiff(
            "The MediumRequest 'a request' has already had "
            "finish_writing called upon it - accept bytes may not be called "
            "anymore.",
            str(error),
        )

    def test_writing_not_completed(self):
        error = errors.WritingNotComplete("a request")
        self.assertEqualDiff(
            "The MediumRequest 'a request' has not has "
            "finish_writing called upon it - until the write phase is complete"
            " no data may be read.",
            str(error),
        )

    def test_transport_not_possible(self):
        error = errors.TransportNotPossible("readonly", "original error")
        self.assertEqualDiff(
            "Transport operation not possible: readonly original error", str(error)
        )

    def assertSocketConnectionError(self, expected, *args, **kwargs):
        """Check the formatting of a SocketConnectionError exception."""
        e = errors.SocketConnectionError(*args, **kwargs)
        self.assertEqual(expected, str(e))

    def test_socket_connection_error(self):
        """Test the formatting of SocketConnectionError."""
        # There should be a default msg about failing to connect
        # we only require a host name.
        self.assertSocketConnectionError("Failed to connect to ahost", "ahost")

        # If port is None, we don't put :None
        self.assertSocketConnectionError(
            "Failed to connect to ahost", "ahost", port=None
        )
        # But if port is supplied we include it
        self.assertSocketConnectionError(
            "Failed to connect to ahost:22", "ahost", port=22
        )

        # We can also supply extra information about the error
        # with or without a port
        self.assertSocketConnectionError(
            "Failed to connect to ahost:22; bogus error",
            "ahost",
            port=22,
            orig_error="bogus error",
        )
        self.assertSocketConnectionError(
            "Failed to connect to ahost; bogus error", "ahost", orig_error="bogus error"
        )
        # An exception object can be passed rather than a string
        orig_error = ValueError("bad value")
        self.assertSocketConnectionError(
            "Failed to connect to ahost; {}".format(str(orig_error)),
            host="ahost",
            orig_error=orig_error,
        )

        # And we can supply a custom failure message
        self.assertSocketConnectionError(
            "Unable to connect to ssh host ahost:444; my_error",
            host="ahost",
            port=444,
            msg="Unable to connect to ssh host",
            orig_error="my_error",
        )

    def test_target_not_branch(self):
        """Test the formatting of TargetNotBranch."""
        error = errors.TargetNotBranch("foo")
        self.assertEqual(
            "Your branch does not have all of the revisions required in "
            "order to merge this merge directive and the target "
            "location specified in the merge directive is not a branch: "
            "foo.",
            str(error),
        )

    def test_unexpected_smart_server_response(self):
        e = errors.UnexpectedSmartServerResponse(("not yes",))
        self.assertEqual(
            "Could not understand response from smart server: ('not yes',)", str(e)
        )

    def test_check_error(self):
        e = errors.BzrCheckError("example check failure")
        self.assertEqual("Internal check failed: example check failure", str(e))
        self.assertTrue(e.internal_error)

    def test_repository_data_stream_error(self):
        """Test the formatting of RepositoryDataStreamError."""
        e = errors.RepositoryDataStreamError("my reason")
        self.assertEqual("Corrupt or incompatible data stream: my reason", str(e))

    def test_immortal_pending_deletion_message(self):
        err = errors.ImmortalPendingDeletion("foo")
        self.assertEqual(
            "Unable to delete transform temporary directory foo.  "
            "Please examine foo to see if it contains any files "
            "you wish to keep, and delete it when you are done.",
            str(err),
        )

    def test_invalid_url_join(self):
        """Test the formatting of InvalidURLJoin."""
        e = urlutils.InvalidURLJoin("Reason", "base path", ("args",))
        self.assertEqual(
            "Invalid URL join request: Reason: 'base path' + ('args',)", str(e)
        )

    def test_unable_encode_path(self):
        err = errors.UnableEncodePath("foo", "executable")
        self.assertEqual(
            "Unable to encode executable path 'foo' in "
            "user encoding " + osutils.get_user_encoding(),
            str(err),
        )

    def test_unknown_format(self):
        err = errors.UnknownFormatError("bar", kind="foo")
        self.assertEqual("Unknown foo format: 'bar'", str(err))

    def test_tip_change_rejected(self):
        err = errors.TipChangeRejected("Unicode message\N{INTERROBANG}")
        self.assertEqual(
            "Tip change rejected: Unicode message\N{INTERROBANG}", str(err)
        )

    def test_error_from_smart_server(self):
        error_tuple = ("error", "tuple")
        err = errors.ErrorFromSmartServer(error_tuple)
        self.assertEqual(
            "Error received from smart server: ('error', 'tuple')", str(err)
        )

    def test_unresumable_write_group(self):
        repo = "dummy repo"
        wg_tokens = ["token"]
        reason = "a reason"
        err = errors.UnresumableWriteGroup(repo, wg_tokens, reason)
        self.assertEqual(
            "Repository dummy repo cannot resume write group ['token']: a reason",
            str(err),
        )

    def test_unsuspendable_write_group(self):
        repo = "dummy repo"
        err = errors.UnsuspendableWriteGroup(repo)
        self.assertEqual(
            "Repository dummy repo cannot suspend a write group.", str(err)
        )

    def test_not_branch_no_args(self):
        err = errors.NotBranchError("path")
        self.assertEqual('Not a branch: "path".', str(err))

    def test_not_branch_bzrdir_with_recursive_not_branch_error(self):
        class FakeBzrDir:
            def open_repository(self):
                # str() on the NotBranchError will trigger a call to this,
                # which in turn will another, identical NotBranchError.
                raise errors.NotBranchError("path", controldir=FakeBzrDir())

        err = errors.NotBranchError("path", controldir=FakeBzrDir())
        self.assertEqual('Not a branch: "path": NotBranchError.', str(err))

    def test_recursive_bind(self):
        error = errors.RecursiveBind("foo_bar_branch")
        msg = (
            'Branch "foo_bar_branch" appears to be bound to itself. '
            "Please use `brz unbind` to fix."
        )
        self.assertEqualDiff(msg, str(error))


class PassThroughError(errors.BzrError):
    _fmt = """Pass through %(foo)s and %(bar)s"""

    def __init__(self, foo, bar):
        errors.BzrError.__init__(self, foo=foo, bar=bar)


class ErrorWithBadFormat(errors.BzrError):
    _fmt = """One format specifier: %(thing)s"""


class ErrorWithNoFormat(errors.BzrError):
    __doc__ = """This class has a docstring but no format string."""


class TestErrorFormatting(tests.TestCase):
    def test_always_str(self):
        e = PassThroughError("\xb5", "bar")
        self.assertIsInstance(e.__str__(), str)
        # In Python 2 str(foo) *must* return a real byte string
        # not a Unicode string. The following line would raise a
        # Unicode error, because it tries to call str() on the string
        # returned from e.__str__(), and it has non ascii characters
        s = str(e)
        self.assertEqual("Pass through \xb5 and bar", s)

    def test_missing_format_string(self):
        e = ErrorWithNoFormat(param="randomvalue")
        self.assertStartsWith(str(e), "Unprintable exception ErrorWithNoFormat")

    def test_mismatched_format_args(self):
        # Even though ErrorWithBadFormat's format string does not match the
        # arguments we constructing it with, we can still stringify an instance
        # of this exception. The resulting string will say its unprintable.
        e = ErrorWithBadFormat(not_thing="x")
        self.assertStartsWith(str(e), "Unprintable exception ErrorWithBadFormat")

    def test_cannot_bind_address(self):
        # see <https://bugs.launchpad.net/bzr/+bug/286871>
        e = errors.CannotBindAddress(
            "example.com", 22, OSError(13, "Permission denied")
        )
        self.assertContainsRe(
            str(e), r'Cannot bind address "example\.com:22":.*Permission denied'
        )


class TestErrorsUsingTransport(tests.TestCaseWithMemoryTransport):
    """Tests for errors that need to use a branch or repo."""

    def test_no_public_branch(self):
        b = self.make_branch(".")
        error = errors.NoPublicBranch(b)
        url = urlutils.unescape_for_display(b.base, "ascii")
        self.assertEqualDiff(
            'There is no public branch set for "{}".'.format(url), str(error)
        )

    def test_no_repo(self):
        dir = controldir.ControlDir.create(self.get_url())
        error = errors.NoRepositoryPresent(dir)
        self.assertNotEqual(-1, str(error).find(dir.transport.clone("..").base))
        self.assertEqual(-1, str(error).find(dir.transport.base))

    def test_corrupt_repository(self):
        repo = self.make_repository(".")
        error = errors.CorruptRepository(repo)
        self.assertEqualDiff(
            "An error has been detected in the repository {}.\n"
            "Please run brz reconcile on this repository.".format(
                repo.controldir.root_transport.base
            ),
            str(error),
        )

    def test_not_branch_bzrdir_with_repo(self):
        controldir = self.make_repository("repo").controldir
        err = errors.NotBranchError("path", controldir=controldir)
        self.assertEqual('Not a branch: "path": location is a repository.', str(err))

    def test_not_branch_bzrdir_without_repo(self):
        controldir = self.make_controldir("bzrdir")
        err = errors.NotBranchError("path", controldir=controldir)
        self.assertEqual('Not a branch: "path".', str(err))

    def test_not_branch_laziness(self):
        real_bzrdir = self.make_controldir("path")

        class FakeBzrDir:
            def __init__(self):
                self.calls = []

            def open_repository(self):
                self.calls.append("open_repository")
                raise errors.NoRepositoryPresent(real_bzrdir)

        fake_bzrdir = FakeBzrDir()
        err = errors.NotBranchError("path", controldir=fake_bzrdir)
        self.assertEqual([], fake_bzrdir.calls)
        str(err)
        self.assertEqual(["open_repository"], fake_bzrdir.calls)
        # Stringifying twice doesn't try to open a repository twice.
        str(err)
        self.assertEqual(["open_repository"], fake_bzrdir.calls)



# Copyright (C) 2005, 2006, 2007, 2009, 2011, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for brz revno."""

import os

from breezy import tests
from breezy.errors import NoSuchRevision


class TestRevno(tests.TestCaseWithTransport):
    def test_revno(self):
        def bzr(*args, **kwargs):
            return self.run_bzr(*args, **kwargs)[0]

        os.mkdir("a")
        os.chdir("a")
        bzr("init")
        self.assertEqual(int(bzr("revno")), 0)

        with open("foo", "wb") as f:
            f.write(b"foo\n")
        bzr("add foo")
        bzr("commit -m foo")
        self.assertEqual(int(bzr("revno")), 1)

        os.mkdir("baz")
        bzr("add baz")
        bzr("commit -m baz")
        self.assertEqual(int(bzr("revno")), 2)

        os.chdir("..")
        self.assertEqual(int(bzr("revno a")), 2)
        self.assertEqual(int(bzr("revno a/baz")), 2)

    def test_revno_tree(self):
        # Make branch and checkout
        wt = self.make_branch_and_tree("branch")
        wt.branch.create_checkout("checkout", lightweight=True)

        # Get the checkout out of date
        self.build_tree(["branch/file"])
        wt.add(["file"])
        wt.commit("mkfile")

        # Make sure revno says we're on 1
        out, err = self.run_bzr("revno checkout")
        self.assertEqual("", err)
        self.assertEqual("1\n", out)

        # Make sure --tree knows it's still on 0
        out, err = self.run_bzr("revno --tree checkout")
        self.assertEqual("", err)
        self.assertEqual("0\n", out)

    def test_revno_tree_no_tree(self):
        # Make treeless branch
        self.make_branch("branch")

        # Try getting it's --tree revno
        out, err = self.run_bzr("revno --tree branch", retcode=3)
        self.assertEqual("", out)
        self.assertEqual('brz: ERROR: No WorkingTree exists for "branch".\n', err)

    def test_dotted_revno_tree(self):
        builder = self.make_branch_builder("branch")
        builder.start_series()
        builder.build_snapshot(
            None,
            [
                ("add", ("", b"root-id", "directory", None)),
                ("add", ("file", b"file-id", "file", b"content\n")),
            ],
            revision_id=b"A-id",
        )
        builder.build_snapshot([b"A-id"], [], revision_id=b"B-id")
        builder.build_snapshot([b"A-id", b"B-id"], [], revision_id=b"C-id")
        builder.finish_series()
        b = builder.get_branch()
        b.create_checkout("checkout_b", lightweight=True, revision_id=b"B-id")
        out, err = self.run_bzr("revno checkout_b")
        self.assertEqual("", err)
        self.assertEqual("2\n", out)
        out, err = self.run_bzr("revno --tree checkout_b")
        self.assertEqual("", err)
        self.assertEqual("1.1.1\n", out)

    def test_stale_revno_tree(self):
        builder = self.make_branch_builder("branch")
        builder.start_series()
        builder.build_snapshot(
            None,
            [
                ("add", ("", b"root-id", "directory", None)),
                ("add", ("file", b"file-id", "file", b"content\n")),
            ],
            revision_id=b"A-id",
        )
        builder.build_snapshot([b"A-id"], [], revision_id=b"B-id")
        builder.build_snapshot([b"A-id"], [], revision_id=b"C-id")
        builder.finish_series()
        b = builder.get_branch()
        # The branch is now at "C-id", but the checkout is still at "B-id"
        # which is no longer in the history
        b.create_checkout("checkout_b", lightweight=True, revision_id=b"B-id")
        out, err = self.run_bzr("revno checkout_b")
        self.assertEqual("", err)
        self.assertEqual("2\n", out)
        out, err = self.run_bzr("revno --tree checkout_b")
        self.assertEqual("", err)
        self.assertEqual("???\n", out)

    def test_revno_ghost(self):
        builder = self.make_branch_builder("branch")
        builder.start_series()
        builder.build_snapshot(
            [b"aghost"],
            [
                ("add", ("", b"root-id", "directory", None)),
                ("add", ("file", b"file-id", "file", b"content\n")),
            ],
            revision_id=b"A-id",
            allow_leftmost_as_ghost=True,
        )
        builder.finish_series()
        b = builder.get_branch()

        def revision_id_to_revno(s, r):
            raise NoSuchRevision(s, r)

        self.overrideAttr(type(b), "revision_id_to_dotted_revno", revision_id_to_revno)
        self.overrideAttr(type(b), "revision_id_to_revno", revision_id_to_revno)
        out, err = self.run_bzr("revno branch")
        self.assertEqual("", err)
        self.assertEqual("???\n", out)

    def test_revno_with_revision(self):
        wt = self.make_branch_and_tree(".")
        revid1 = wt.commit("rev1")
        wt.commit("rev2")

        out, err = self.run_bzr("revno -r-2 .")
        self.assertEqual("1\n", out)

        out, err = self.run_bzr("revno -rrevid:{} .".format(revid1.decode("utf-8")))
        self.assertEqual("1\n", out)

    def test_revno_and_tree_mutually_exclusive(self):
        self.make_branch_and_tree(".")
        out, err = self.run_bzr("revno -r-2 --tree .", retcode=3)
        self.assertEqual("", out)
        self.assertEqual(
            "brz: ERROR: --tree and --revision can not be used together\n", err
        )



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for interface conformance of 'WorkingTree.rename_one'."""

import os

from breezy import errors, osutils, tests
from breezy import transport as _mod_transport
from breezy.tests import features
from breezy.tests.matchers import HasPathRelations
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestRenameOne(TestCaseWithWorkingTree):
    def assertPathRelations(self, previous_tree, tree, relations):
        self.assertThat(tree, HasPathRelations(previous_tree, relations))

    def test_rename_one_target_not_dir(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        tree.add(["a"])
        tree.commit("initial")

        self.assertRaises(
            errors.BzrMoveFailedError, tree.rename_one, "a", "not-a-dir/b"
        )

    def test_rename_one_non_existent(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/"])
        tree.add(["a"])
        tree.commit("initial")
        self.assertRaises(
            errors.BzrMoveFailedError, tree.rename_one, "not-a-file", "a/failure"
        )
        self.assertRaises(
            errors.BzrMoveFailedError, tree.rename_one, "not-a-file", "also_not"
        )

    def test_rename_one_target_not_versioned(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "b"])
        tree.add(["b"])
        tree.commit("initial")
        if tree.has_versioned_directories():
            self.assertRaises(errors.BzrMoveFailedError, tree.rename_one, "b", "a/b")
        else:
            tree.rename_one("b", "a/b")

    def test_rename_one_unversioned(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "b"])
        tree.add(["a"])
        tree.commit("initial")
        self.assertRaises(errors.BzrMoveFailedError, tree.rename_one, "b", "a/b")

    def test_rename_one_samedir(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/"])
        tree.add(["a", "b"])
        tree.commit("initial")

        a_contents = tree.get_file_text("a")
        tree.rename_one("a", "foo")
        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("b/", "b/"), ("foo", "a")]
        )
        self.assertPathDoesNotExist("a")
        self.assertFileEqual(a_contents, "foo")

    def test_rename_one_not_localdir(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a", "tree/b/"])
        tree.add(["a", "b"])
        tree.commit("initial")

        a_contents = tree.get_file_text("a")
        tree.rename_one("a", "b/foo")
        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("b/", "b/"), ("b/foo", "a")]
        )
        self.assertPathDoesNotExist("tree/a")
        self.assertFileEqual(a_contents, "tree/b/foo")

    def test_rename_one_subdir(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/", "b/c"])
        tree.add(["a", "b", "b/c"])
        tree.commit("initial")
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [("", ""), ("a", "a"), ("b/", "b/"), ("b/c", "b/c")],
        )
        a_contents = tree.get_file_text("a")
        tree.rename_one("a", "b/d")
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [("", ""), ("b/", "b/"), ("b/c", "b/c"), ("b/d", "a")],
        )
        self.assertPathDoesNotExist("a")
        self.assertFileEqual(a_contents, "b/d")

    def test_rename_one_parent_dir(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/", "b/c"])
        tree.add(["a", "b", "b/c"])
        tree.commit("initial")
        c_contents = tree.get_file_text("b/c")
        tree.rename_one("b/c", "d")
        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("a", "a"), ("b/", "b/"), ("d", "b/c")]
        )
        self.assertPathDoesNotExist("b/c")
        self.assertFileEqual(c_contents, "d")

    def test_rename_one_fail_consistent(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/", "b/a", "c"])
        tree.add(["a", "b", "c"])
        tree.commit("initial")
        # Target already exists
        self.assertRaises(errors.RenameFailedFilesExist, tree.rename_one, "a", "b/a")
        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("a", "a"), ("b/", "b/"), ("c", "c")]
        )

    def test_rename_one_onto_existing(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b"])
        tree.add(["a", "b"])
        tree.commit("initial")

        self.assertRaises(errors.BzrMoveFailedError, tree.rename_one, "a", "b")

    def test_rename_one_onto_self(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["b/", "b/a"])
        tree.add(["b", "b/a"])
        tree.commit("initial")

        self.assertRaises(errors.BzrMoveFailedError, tree.rename_one, "b/a", "b/a")

    def test_rename_one_onto_self_root(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        tree.add(["a"])
        tree.commit("initial")

        self.assertRaises(errors.BzrMoveFailedError, tree.rename_one, "a", "a")

    def test_rename_one_after(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/"])
        tree.add(["a", "b"])
        tree.commit("initial")
        os.rename("a", "b/foo")

        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("a", "a"), ("b/", "b/")]
        )

        # We don't need after=True as long as source is missing and target
        # exists.
        tree.rename_one("a", "b/foo")
        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("b/", "b/"), ("b/foo", "a")]
        )

    def test_rename_one_after_with_after(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/"])
        tree.add(["a", "b"])
        tree.commit("initial")
        os.rename("a", "b/foo")

        if tree.has_versioned_directories():
            self.assertPathRelations(
                tree.basis_tree(), tree, [("", ""), ("a", "a"), ("b/", "b/")]
            )
        else:
            self.assertPathRelations(tree.basis_tree(), tree, [("", ""), ("a", "a")])

        # Passing after=True should work as well
        tree.rename_one("a", "b/foo", after=True)
        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("b/", "b/"), ("b/foo", "a")]
        )

    def test_rename_one_after_dest_versioned(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        tree.add(["a"])
        tree.commit("initial")
        os.rename("a", "b")
        tree.add(["b"])

        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("a", "a"), ("b", None)]
        )

        e = self.assertRaises(errors.BzrMoveFailedError, tree.rename_one, "a", "b")
        self.assertIsInstance(e.extra, errors.AlreadyVersionedError)

    def test_rename_one_after_with_after_dest_versioned(self):
        """Using after with an already versioned file should fail."""
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b"])
        tree.add(["a", "b"])
        tree.commit("initial")
        os.unlink("a")

        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("a", "a"), ("b", "b")]
        )

        e = self.assertRaises(
            errors.BzrMoveFailedError, tree.rename_one, "a", "b", after=True
        )
        self.assertIsInstance(e.extra, errors.AlreadyVersionedError)

    def test_rename_one_after_with_after_dest_added(self):
        """Using after with a newly added file should work."""
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        tree.add(["a"])
        tree.commit("initial")
        os.rename("a", "b")
        tree.add(["b"])

        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("a", "a"), ("b", None)]
        )
        tree.rename_one("a", "b", after=True)
        self.assertPathRelations(tree.basis_tree(), tree, [("", ""), ("b", "a")])

    def test_rename_one_after_source_removed(self):
        """Rename even if the source was already unversioned."""
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/"])
        tree.add(["a", "b"])
        tree.commit("initial")
        os.rename("a", "b/foo")
        tree.remove(["a"])

        self.assertPathRelations(tree.basis_tree(), tree, [("", ""), ("b/", "b/")])

        # We don't need after=True as long as source is missing and target
        # exists.
        tree.rename_one("a", "b/foo")
        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("b/", "b/"), ("b/foo", "a")]
        )

    def test_rename_one_after_no_target(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/"])
        tree.add(["a", "b"])
        tree.commit("initial")

        # Passing after when the file hasn't been rename_one raises an
        # exception
        self.assertRaises(
            errors.BzrMoveFailedError, tree.rename_one, "a", "b/foo", after=True
        )
        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("a", "a"), ("b/", "b/")]
        )

    def test_rename_one_after_source_and_dest(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/", "b/foo"])
        tree.add(["a", "b"])
        tree.commit("initial")

        # TODO: jam 20070225 I would usually use 'rb', but assertFileEqual
        #       uses 'r'.
        with open("a") as a_file:
            a_text = a_file.read()
        with open("b/foo") as foo_file:
            foo_text = foo_file.read()

        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("a", "a"), ("b/", "b/")]
        )
        self.assertRaises(
            errors.RenameFailedFilesExist, tree.rename_one, "a", "b/foo", after=False
        )
        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("a", "a"), ("b/", "b/")]
        )
        self.assertFileEqual(a_text, "a")
        self.assertFileEqual(foo_text, "b/foo")
        # But you can pass after=True
        tree.rename_one("a", "b/foo", after=True)
        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("b/", "b/"), ("b/foo", "a")]
        )
        # But it shouldn't actually move anything
        self.assertFileEqual(a_text, "a")
        self.assertFileEqual(foo_text, "b/foo")

    def test_rename_one_directory(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b", "a/c/", "a/c/d", "e/"])
        tree.add(["a", "a/b", "a/c", "a/c/d", "e"])
        tree.commit("initial")

        tree.rename_one("a", "e/f")
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [
                ("", ""),
                ("e/", "e/"),
                ("e/f/", "a/"),
                ("e/f/b", "a/b"),
                ("e/f/c/", "a/c/"),
                ("e/f/c/d", "a/c/d"),
            ],
        )

    def test_rename_one_moved(self):
        """Moving a moved entry works as expected."""
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b", "c/"])
        tree.add(["a", "a/b", "c"])
        tree.commit("initial")

        tree.rename_one("a/b", "c/foo")
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [("", ""), ("a/", "a/"), ("c/", "c/"), ("c/foo", "a/b")],
        )

        tree.rename_one("c/foo", "bar")
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [("", ""), ("a/", "a/"), ("bar", "a/b"), ("c/", "c/")],
        )

    def test_rename_to_denormalised_fails(self):
        if osutils.normalizes_filenames():
            raise tests.TestNotApplicable("OSX normalizes filenames")
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        tree.add(["a"])
        self.assertRaises(
            (errors.InvalidNormalization, UnicodeEncodeError),
            tree.rename_one,
            "a",
            "ba\u030arry",
        )

    def test_rename_unversioned_non_ascii(self):
        """Check error when renaming an unversioned non-ascii file."""
        self.requireFeature(features.UnicodeFilenameFeature)
        tree = self.make_branch_and_tree(".")
        self.build_tree(["\xa7"])
        e = self.assertRaises(errors.BzrRenameFailedError, tree.rename_one, "\xa7", "b")
        self.assertIsInstance(e.extra, errors.NotVersionedError)
        self.assertEqual(e.extra.path, "\xa7")

    def test_rename_into_unversioned_non_ascii_dir(self):
        """Check error when renaming into unversioned non-ascii directory."""
        self.requireFeature(features.UnicodeFilenameFeature)
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "\xa7/"])
        tree.add(["a"])
        if tree.has_versioned_directories():
            e = self.assertRaises(
                errors.BzrMoveFailedError, tree.rename_one, "a", "\xa7/a"
            )
            self.assertIsInstance(e.extra, errors.NotVersionedError)
            self.assertEqual(e.extra.path, "\xa7")
        else:
            tree.rename_one("a", "\xa7/a")

    def test_rename_over_already_versioned_non_ascii(self):
        """Check error renaming over an already versioned non-ascii file."""
        self.requireFeature(features.UnicodeFilenameFeature)
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "\xa7"])
        tree.add(["a", "\xa7"])
        e = self.assertRaises(errors.BzrMoveFailedError, tree.rename_one, "a", "\xa7")
        self.assertIsInstance(e.extra, errors.AlreadyVersionedError)
        self.assertEqual(e.extra.path, "\xa7")

    def test_rename_after_non_existant_non_ascii(self):
        """Check error renaming after move with missing non-ascii file."""
        self.requireFeature(features.UnicodeFilenameFeature)
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        tree.add(["a"])
        e = self.assertRaises(
            errors.BzrMoveFailedError, tree.rename_one, "a", "\xa7", after=True
        )
        self.assertIsInstance(e.extra, _mod_transport.NoSuchFile)
        self.assertEqual(e.extra.path, "\xa7")



#!/usr/bin/python3

# Simple script that will output the release where a given bug was fixed
# searching the NEWS file

import optparse
import re
import sys


class NewsParser:
    paren_exp_re = re.compile(r"\(([^)]+)\)")
    release_re = re.compile("brz[ -]")
    release_prefix_length = len("brz ")
    bugs_re = re.compile("#([0-9]+)")

    def __init__(self, news):
        self.news = news
        # Temporary attributes used by the parser
        self.release = None
        self.date = None
        self.may_be_release = None
        self.release_markup = None
        self.entry = ""
        self.line = None
        self.lrs = None

    def set_line(self, line):
        self.line = line
        self.lrs = line.rstrip()

    def try_release(self):
        if self.release_re.match(self.lrs) is not None:
            # May be a new release
            self.may_be_release = self.lrs
            # We know the markup will have the same length as the release
            self.release_markup = "#" * len(self.may_be_release)
            return True
        return False

    def confirm_release(self):
        if self.may_be_release is not None and self.lrs == self.release_markup:
            # The release is followed by the right markup
            self.release = self.may_be_release[self.release_prefix_length :]
            # Wait for the associated date
            self.date = None
            return True
        return False

    def try_date(self):
        if self.release is None:
            return False
        date_re = re.compile(
            r":{}: (NOT RELEASED YET|\d{{4}}-\d{{2}}-\d{{2}})".format(self.release)
        )
        match = date_re.match(self.lrs)
        if match is not None:
            self.date = match.group(1)
            return True
        # The old fashion way
        released_re = re.compile(r":Released:\s+(\d{4}-\d{2}-\d{2})")
        match = released_re.match(self.lrs)
        if match is not None:
            self.date = match.group(1)
            return True
        return False

    def add_line_to_entry(self):
        if self.lrs == "":
            return False
        self.entry += self.line
        return True

    def extract_bugs_from_entry(self):
        """Possibly extract bugs from a NEWS entry and yield them.

        Not all entries will contain bugs and some entries are even garbage and
        we don't try to parse them (yet). The trigger is a '#' and what looks
        like a bug number inside parens to start with. From that we extract
        authors (when present) and multiple bugs if needed.
        """
        # FIXME: Malone entries are different
        # Join all entry lines to simplify multiple line matching
        flat_entry = " ".join(self.entry.splitlines())
        # Fixed bugs are always inside parens
        for par in self.paren_exp_re.findall(flat_entry):
            sharp = par.find("#")
            if sharp is not None:
                # We have at least one bug inside parens.
                bugs = list(self.bugs_re.finditer(par))
                if bugs:
                    # See where the first bug is mentioned
                    start = bugs[0].start()
                    end = bugs[-1].end()
                    if start == 0:
                        # (bugs/authors)
                        authors = par[end:]
                    else:
                        # (authors/bugs)
                        authors = par[:start]
                    for bug_match in bugs:
                        bug_number = bug_match.group(0)
                        yield (bug_number, authors, self.release, self.date, self.entry)
        # We've consumed the entry
        self.entry = ""

    def parse_bugs(self):
        for line in self.news:
            self.set_line(line)
            if self.try_release():
                continue  # line may a be release
            try:
                if self.confirm_release():
                    continue  # previous line was indeed a release
            finally:
                self.may_be_release = None
            if self.try_date():
                continue  # The release date has been seen
            if self.add_line_to_entry():
                continue  # accumulate in self.enrty
            yield from self.extract_bugs_from_entry()


def main():
    opt_parser = optparse.OptionParser(
        usage="""Usage: %prog [options] BUG_NUMBER
    """
    )
    opt_parser.add_option(
        "-f",
        "--file",
        type="str",
        dest="news_file",
        help="NEWS file (defaults to ./NEWS)",
    )
    opt_parser.add_option(
        "-m",
        "--message",
        type="str",
        dest="msg_re",
        help="A regexp to search for in the news entry "
        "(BUG_NUMBER should not be specified in this case)",
    )
    opt_parser.set_defaults(news_file="./NEWS")
    (opts, args) = opt_parser.parse_args(sys.argv[1:])
    if opts.msg_re is not None:
        if len(args) != 0:
            opt_parser.error("BUG_NUMBER and -m are mutually exclusive")
        bug = None
        msg_re = re.compile(opts.msg_re)
    elif len(args) != 1:
        opt_parser.error("Expected a single bug number, got {!r}".format(args))
    else:
        bug = args[0]

    news = open(opts.news_file)
    parser = NewsParser(news)
    try:
        seen = 0
        for b in parser.parse_bugs():
            (
                number,
                authors,
                release,
                date,
                entry,
            ) = b
            # indent entry
            entry = "\n".join(["    " + l for l in entry.splitlines()])
            found = False
            if bug is not None:
                if number[1:] == bug:  # Strip the leading '#'
                    found = True
            elif msg_re.search(entry) is not None:
                found = True
            if found:
                print(
                    "Bug {} was fixed in brz-{}/{} by {}:".format(
                        number, release, date, authors
                    )
                )
                print(entry)
            seen += 1
    finally:
        print("{} bugs seen".format(seen))
        news.close()


main()



# Copyright (C) 2008, 2009, 2010, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests that an enabled view is reported and impacts expected commands."""

from breezy import osutils, tests


class TestViewFileOperations(tests.TestCaseWithTransport):
    def make_abc_tree_with_ab_view(self):
        wt = self.make_branch_and_tree(".")
        self.build_tree(["a", "b", "c"])
        wt.views.set_view("my", ["a", "b"])
        return wt

    def test_view_on_status(self):
        self.make_abc_tree_with_ab_view()
        out, err = self.run_bzr("status")
        self.assertEqual("Ignoring files outside view. View is a, b\n", err)
        self.assertEqual("unknown:\n  a\n  b\n", out)

    def test_view_on_status_selected(self):
        self.make_abc_tree_with_ab_view()
        out, err = self.run_bzr("status a")
        self.assertEqual("", err)
        self.assertEqual("unknown:\n  a\n", out)
        out, err = self.run_bzr("status c", retcode=3)
        self.assertEqual(
            'brz: ERROR: Specified file "c" is outside the current view: a, b\n', err
        )
        self.assertEqual("", out)

    def test_view_on_add(self):
        self.make_abc_tree_with_ab_view()
        out, err = self.run_bzr("add")
        self.assertEqual("Ignoring files outside view. View is a, b\n", err)
        self.assertEqual("adding a\nadding b\n", out)

    def test_view_on_add_selected(self):
        self.make_abc_tree_with_ab_view()
        out, err = self.run_bzr("add a")
        self.assertEqual("", err)
        self.assertEqual("adding a\n", out)
        out, err = self.run_bzr("add c", retcode=3)
        self.assertEqual(
            'brz: ERROR: Specified file "c" is outside the current view: a, b\n', err
        )
        self.assertEqual("", out)

    def test_view_on_diff(self):
        self.make_abc_tree_with_ab_view()
        self.run_bzr("add")
        out, err = self.run_bzr("diff", retcode=1)
        self.assertEqual("*** Ignoring files outside view. View is a, b\n", err)

    def test_view_on_diff_selected(self):
        self.make_abc_tree_with_ab_view()
        self.run_bzr("add")
        out, err = self.run_bzr("diff a", retcode=1)
        self.assertEqual("", err)
        self.assertStartsWith(out, "=== added file 'a'\n")
        out, err = self.run_bzr("diff c", retcode=3)
        self.assertEqual(
            'brz: ERROR: Specified file "c" is outside the current view: a, b\n', err
        )
        self.assertEqual("", out)

    def test_view_on_commit(self):
        self.make_abc_tree_with_ab_view()
        self.run_bzr("add")
        out, err = self.run_bzr('commit -m "testing commit"')
        err_lines = err.splitlines()
        self.assertEqual("Ignoring files outside view. View is a, b", err_lines[0])
        self.assertStartsWith(err_lines[1], "Committing to:")
        # TODO(jelmer): Should 'added ' output for commit be sorted?
        self.assertIn("added a", [err_lines[2], err_lines[3]])
        self.assertIn("added b", [err_lines[2], err_lines[3]])
        self.assertEqual("Committed revision 1.", err_lines[4])
        self.assertEqual("", out)

    def test_view_on_commit_selected(self):
        self.make_abc_tree_with_ab_view()
        self.run_bzr("add")
        out, err = self.run_bzr('commit -m "file in view" a')
        err_lines = err.splitlines()
        self.assertStartsWith(err_lines[0], "Committing to:")
        self.assertEqual("added a", err_lines[1])
        self.assertEqual("Committed revision 1.", err_lines[2])
        self.assertEqual("", out)
        out, err = self.run_bzr('commit -m "file out of view" c', retcode=3)
        self.assertEqual(
            'brz: ERROR: Specified file "c" is outside the current view: a, b\n', err
        )
        self.assertEqual("", out)

    def test_view_on_remove_selected(self):
        self.make_abc_tree_with_ab_view()
        self.run_bzr("add")
        out, err = self.run_bzr("remove --keep a")
        self.assertEqual("removed a\n", err)
        self.assertEqual("", out)
        out, err = self.run_bzr("remove --keep c", retcode=3)
        self.assertEqual(
            'brz: ERROR: Specified file "c" is outside the current view: a, b\n', err
        )
        self.assertEqual("", out)

    def test_view_on_revert(self):
        self.make_abc_tree_with_ab_view()
        self.run_bzr("add")
        out, err = self.run_bzr("revert")
        err_lines = err.splitlines()
        self.assertEqual("Ignoring files outside view. View is a, b", err_lines[0])
        self.assertEqual("-   a", err_lines[1])
        self.assertEqual("-   b", err_lines[2])
        self.assertEqual("", out)

    def test_view_on_revert_selected(self):
        self.make_abc_tree_with_ab_view()
        self.run_bzr("add")
        out, err = self.run_bzr("revert a")
        self.assertEqual("-   a\n", err)
        self.assertEqual("", out)
        out, err = self.run_bzr("revert c", retcode=3)
        self.assertEqual(
            'brz: ERROR: Specified file "c" is outside the current view: a, b\n', err
        )
        self.assertEqual("", out)

    def test_view_on_ls(self):
        self.make_abc_tree_with_ab_view()
        self.run_bzr("add")
        out, err = self.run_bzr("ls")
        out_lines = out.splitlines()
        self.assertEqual("Ignoring files outside view. View is a, b\n", err)
        self.assertEqual("a", out_lines[0])
        self.assertEqual("b", out_lines[1])


class TestViewTreeOperations(tests.TestCaseWithTransport):
    def make_abc_tree_and_clone_with_ab_view(self):
        # Build the first tree
        wt1 = self.make_branch_and_tree("tree_1")
        self.build_tree(["tree_1/a", "tree_1/b", "tree_1/c"])
        wt1.add(["a", "b", "c"])
        wt1.commit("adding a b c")
        # Build the second tree and give it a view
        wt2 = wt1.controldir.sprout("tree_2").open_workingtree()
        wt2.views.set_view("my", ["a", "b"])
        # Commit a change to the first tree
        self.build_tree_contents(
            [
                ("tree_1/a", b"changed a\n"),
                ("tree_1/c", b"changed c\n"),
            ]
        )
        wt1.commit("changing a c")
        return wt1, wt2

    def test_view_on_pull(self):
        tree_1, tree_2 = self.make_abc_tree_and_clone_with_ab_view()
        out, err = self.run_bzr("pull -d tree_2 tree_1")
        self.assertEqualDiff(
            "Operating on whole tree but only reporting on 'my' view.\n"
            " M  a\n"
            "All changes applied successfully.\n",
            err,
        )
        self.assertEqualDiff("Now on revision 2.\n", out)

    def test_view_on_update(self):
        tree_1, tree_2 = self.make_abc_tree_and_clone_with_ab_view()
        self.run_bzr("bind ../tree_1", working_dir="tree_2")
        out, err = self.run_bzr("update", working_dir="tree_2")
        self.assertEqualDiff(
            """Operating on whole tree but only reporting on 'my' view.
 M  a
All changes applied successfully.
Updated to revision 2 of branch {}
""".format(osutils.pathjoin(self.test_dir, "tree_1")),
            err,
        )
        self.assertEqual("", out)

    def test_view_on_merge(self):
        tree_1, tree_2 = self.make_abc_tree_and_clone_with_ab_view()
        out, err = self.run_bzr("merge -d tree_2 tree_1")
        self.assertEqualDiff(
            "Operating on whole tree but only reporting on 'my' view.\n"
            " M  a\n"
            "All changes applied successfully.\n",
            err,
        )
        self.assertEqual("", out)



# Copyright (C) 2019 Breezy Developers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from breezy.tests import TestCaseWithTransport


class TestPatch(TestCaseWithTransport):
    def test_patch(self):
        self.run_bzr("init")
        with open("myfile", "w") as f:
            f.write("hello")
        self.run_bzr("add")
        self.run_bzr("commit -m hello")
        with open("myfile", "w") as f:
            f.write("goodbye")
        with open("mypatch", "w") as f:
            f.write(self.run_bzr("diff -p1", retcode=1)[0])
        self.run_bzr("revert")
        self.assertFileEqual("hello", "myfile")
        self.run_bzr("patch -p1 --silent mypatch")
        self.assertFileEqual("goodbye", "myfile")

    def test_patch_invalid_strip(self):
        self.run_bzr_error(
            args="patch --strip=a",
            error_regexes=["brz: ERROR: invalid value for option -p/--strip: a"],
        )



# Copyright (C) 2010, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Tests for generating multiple tests for scenarios."""

from breezy.tests import TestCase, TestLoader, iter_suite_tests, multiply_tests
from breezy.tests.scenarios import (
    load_tests_apply_scenarios,
    multiply_scenarios,
    multiply_tests_by_their_scenarios,
)

# There aren't any actually parameterized tests here, but this exists as a
# demonstration; so that you can interactively observe them being multiplied;
# and so that we check everything hooks up properly.
load_tests = load_tests_apply_scenarios


def vary_by_color():
    """Very simple static variation example."""
    for color in ["red", "green", "blue"]:
        yield (color, {"color": color})


def vary_named_attribute(attr_name):
    """More sophisticated: vary a named parameter."""
    yield ("a", {attr_name: "a"})
    yield ("b", {attr_name: "b"})


def get_generated_test_attributes(suite, attr_name):
    """Return the `attr_name` attribute from all tests in the suite."""
    return sorted([getattr(t, attr_name) for t in iter_suite_tests(suite)])


class TestTestScenarios(TestCase):
    def test_multiply_tests(self):
        loader = TestLoader()
        suite = loader.suiteClass()
        multiply_tests(self, vary_by_color(), suite)
        self.assertEqual(
            ["blue", "green", "red"], get_generated_test_attributes(suite, "color")
        )

    def test_multiply_scenarios_from_generators(self):
        """It's safe to multiply scenarios that come from generators."""
        s = multiply_scenarios(
            vary_named_attribute("one"),
            vary_named_attribute("two"),
        )
        self.assertEqual(2 * 2, len(s), s)

    def test_multiply_tests_by_their_scenarios(self):
        loader = TestLoader()
        suite = loader.suiteClass()
        test_instance = PretendVaryingTest("test_nothing")
        multiply_tests_by_their_scenarios(test_instance, suite)
        self.assertEqual(
            ["a", "a", "b", "b"], get_generated_test_attributes(suite, "value")
        )

    def test_multiply_tests_no_scenarios(self):
        """Tests with no scenarios attribute aren't multiplied."""
        suite = TestLoader().suiteClass()
        multiply_tests_by_their_scenarios(self, suite)
        self.assertLength(1, list(iter_suite_tests(suite)))


class PretendVaryingTest(TestCase):
    scenarios = multiply_scenarios(
        vary_named_attribute("value"),
        vary_named_attribute("other"),
    )

    def test_nothing(self):
        """This test exists just so it can be multiplied."""
        pass
